import time
from unittest.mock import MagicMock
from collections import namedtuple
import logging
import pytest

from qtpy.QtCore import Qt
from pyqtgraph import PlotItem, PlotWidget
import numpy as np

from haven import tiled_client
from firefly.main_window import PlanMainWindow
from firefly.run_browser import RunBrowserDisplay
from firefly.run_client import DatabaseWorker


log = logging.getLogger(__name__)


def wait_for_runs_model(display, qtbot):
    with qtbot.waitSignal(display.runs_model_changed):
        pass


@pytest.fixture()
def client(sim_tiled):
    return sim_tiled["255id_testing"]


@pytest.fixture()
def display(client, qtbot, ffapp):
    display = RunBrowserDisplay(root_node=client)
    wait_for_runs_model(display, qtbot)
    yield display
    display._thread.quit()


def test_run_viewer_action(ffapp, monkeypatch, sim_tiled):
    monkeypatch.setattr(ffapp, "create_window", MagicMock())
    assert hasattr(ffapp, "show_run_browser_action")
    ffapp.show_run_browser_action.trigger()
    assert isinstance(ffapp.windows["run_browser"], MagicMock)


def test_load_runs(display):
    assert display.runs_model.rowCount() > 0
    assert display.ui.runs_total_label.text() == str(display.runs_model.rowCount())


def test_update_selected_runs(qtbot, display):
    # Change the proposal item
    selection_model = display.ui.run_tableview.selectionModel()
    item = display.runs_model.item(0, 1)
    assert item is not None
    rect = display.run_tableview.visualRect(item.index())
    with qtbot.waitSignal(display._db_worker.selected_runs_changed):
        qtbot.mouseClick(
            display.run_tableview.viewport(), Qt.LeftButton, pos=rect.center()
        )
    # Check that the runs were saved
    assert len(display._db_worker.selected_runs) > 0


def test_metadata(qtbot, display):
    # Change the proposal item
    selection_model = display.ui.run_tableview.selectionModel()
    item = display.runs_model.item(0, 1)
    assert item is not None
    rect = display.run_tableview.visualRect(item.index())
    with qtbot.waitSignal(display._db_worker.selected_runs_changed):
        qtbot.mouseClick(
            display.run_tableview.viewport(), Qt.LeftButton, pos=rect.center()
        )
    # Check that the metadata was set properly in the Metadata tab
    metadata_doc = display.ui.metadata_textedit.document()
    text = display.ui.metadata_textedit.document().toPlainText()
    assert "xafs_scan" in text


def test_1d_plot_signals(client, display):
    # Check that the 1D plot was created
    plot_widget = display.ui.plot_1d_view
    plot_item = display.plot_1d_item
    assert isinstance(plot_widget, PlotWidget)
    assert isinstance(plot_item, PlotItem)
    # Update the list of runs and see if the controsl get updated
    display._db_worker.selected_runs = client.values()
    display._db_worker.selected_runs_changed.emit([])
    # Check signals in checkboxes
    for combobox in [
        display.ui.multi_signal_x_combobox,
        display.ui.signal_y_combobox,
        display.ui.signal_r_combobox,
        display.ui.signal_x_combobox,
    ]:
        assert (
            combobox.findText("energy_energy") > -1
        ), f"energy_energy signal not in {combobox.objectName()}."


def test_1d_plot_signal_memory(client, display):
    """Do we remember the signals that were previously selected."""
    # Check that the 1D plot was created
    plot_widget = display.ui.plot_1d_view
    plot_item = display.plot_1d_item
    assert isinstance(plot_widget, PlotWidget)
    assert isinstance(plot_item, PlotItem)
    # Update the list of runs and see if the controsl get updated
    display._db_worker.selected_runs = client.values()
    display.update_1d_signals()
    # Check signals in comboboxes
    cb = display.ui.signal_y_combobox
    assert cb.currentText() == "energy_energy"
    cb.setCurrentIndex(1)
    assert cb.currentText() == "energy_id_energy_readback"
    # Update the combobox signals and make sure the text didn't change
    display.update_1d_signals()
    assert cb.currentText() == "energy_id_energy_readback"


def test_1d_hinted_signals(client, display):
    display.ui.plot_1d_hints_checkbox.setChecked(True)
    # Check that the 1D plot was created
    plot_widget = display.ui.plot_1d_view
    plot_item = display.plot_1d_item
    assert isinstance(plot_widget, PlotWidget)
    assert isinstance(plot_item, PlotItem)
    # Update the list of runs and see if the controsl get updated
    display._db_worker.selected_runs = client.values()
    display.update_1d_signals()
    # Check signals in checkboxes
    combobox = display.ui.signal_x_combobox
    assert (
        combobox.findText("energy_energy") > -1
    ), f"hinted signal not in {combobox.objectName()}."
    assert (
        combobox.findText("It_net_counts") == -1
    ), f"unhinted signal found in {combobox.objectName()}."


def test_update_1d_plot(client, display, qtbot):
    run = client.values()[0]
    run_data = run["primary"]["data"].read()
    expected_xdata = run_data.energy_energy
    expected_ydata = np.log(run_data.I0_net_counts / run_data.It_net_counts)
    expected_ydata = np.gradient(expected_ydata, expected_xdata)
    with qtbot.waitSignal(display.plot_1d_changed):
        display._db_worker.selected_runs_changed.emit([])
    # Set the controls to describe the data we want to test
    x_combobox = display.ui.signal_x_combobox
    x_combobox.addItem("energy_energy")
    x_combobox.setCurrentText("energy_energy")
    y_combobox = display.ui.signal_y_combobox
    y_combobox.addItem("It_net_counts")
    y_combobox.setCurrentText("It_net_counts")
    r_combobox = display.ui.signal_r_combobox
    r_combobox.addItem("I0_net_counts")
    r_combobox.setCurrentText("I0_net_counts")
    display.ui.signal_r_checkbox.setChecked(True)
    display.ui.logarithm_checkbox.setChecked(True)
    display.ui.invert_checkbox.setChecked(True)
    display.ui.gradient_checkbox.setChecked(True)
    # Update the plots
    display._db_worker.selected_runs = [run]
    display.update_1d_plot()
    # Check that the data were added
    data_item = display.plot_1d_item.listDataItems()[0]
    xdata, ydata = data_item.getData()
    np.testing.assert_almost_equal(xdata, expected_xdata)
    np.testing.assert_almost_equal(ydata, expected_ydata)


def test_update_multi_plot(client, display, qtbot):
    print("Current text", display.ui.multi_signal_x_combobox.currentText())
    run = client.values()[0]
    run_data = run["primary"]["data"].read()
    expected_xdata = run_data.energy_energy
    expected_ydata = np.log(run_data.I0_net_counts / run_data.It_net_counts)
    expected_ydata = np.gradient(expected_ydata, expected_xdata)
    with qtbot.waitSignal(display.plot_1d_changed):
        display._db_worker.selected_runs_changed.emit([])
    # Configure signals
    display.ui.multi_signal_x_combobox.addItem("energy_energy")
    display.ui.multi_signal_x_combobox.setCurrentText("energy_energy")
    display.multi_y_signals = ["energy_energy"]
    display._db_worker.selected_runs = [run]
    # Update the plots
    display.update_multi_plot()
    # Check that the data were added
    # data_item = display._multiplot_items[0].listDataItems()[0]
    # xdata, ydata = data_item.getData()
    # np.testing.assert_almost_equal(xdata, expected_xdata)
    # np.testing.assert_almost_equal(ydata, expected_ydata)


def test_filter_controls(client, display, qtbot):
    # Does editing text change the filters?
    display.ui.filter_user_combobox.setCurrentText("")
    with qtbot.waitSignal(display.filters_changed):
        qtbot.keyClicks(display.ui.filter_user_combobox, "wolfman")
    # Set some values for the rest of the controls
    display.ui.filter_proposal_combobox.setCurrentText("12345")
    display.ui.filter_esaf_combobox.setCurrentText("678901")
    display.ui.filter_current_proposal_checkbox.setChecked(True)
    display.ui.filter_current_esaf_checkbox.setChecked(True)
    display.ui.filter_plan_combobox.addItem("cake")
    display.ui.filter_plan_combobox.setCurrentText("cake")
    display.ui.filter_full_text_lineedit.setText("Aperature Science")
    display.ui.filter_edge_combobox.setCurrentText("U-K")
    display.ui.filter_sample_combobox.setCurrentText("Pb.*")
    with qtbot.waitSignal(display.filters_changed) as blocker:
        display.update_filters()
    # Check if the filters were update correctly
    filters = blocker.args[0]
    assert filters == {
        "user": "wolfman",
        "proposal": "12345",
        "esaf": "678901",
        "use_current_proposal": True,
        "use_current_esaf": True,
        "exit_status": "success",
        "plan": "cake",
        "full_text": "Aperature Science",
        "edge": "U-K",
        "sample": "Pb.*",
    }


def test_filter_runs(client, qtbot):
    worker = DatabaseWorker(root_node=client)
    worker._filters["plan"] = "xafs_scan"
    with qtbot.waitSignal(worker.all_runs_changed) as blocker:
        worker.load_all_runs()
    # Check that the runs were filtered
    runs = blocker.args[0]
    assert len(runs) == 1


def test_distinct_fields(client, qtbot, display):
    worker = DatabaseWorker(root_node=client)
    with qtbot.waitSignal(worker.distinct_fields_changed) as blocker:
        worker.load_distinct_fields()
    # Check that the dictionary has the right structure
    distinct_fields = blocker.args[0]
    for key in ["sample_name"]:
        assert key in distinct_fields.keys()



import logging

import haven

from firefly import display, FireflyApplication

log = logging.getLogger(__name__)


class StatusDisplay(display.FireflyDisplay):
    def customize_ui(self):
        app = FireflyApplication.instance()
        self.ui.bss_modify_button.clicked.connect(app.show_bss_window_action.trigger)

    def ui_filename(self):
        return "status.ui"



#!/usr/bin/env python3
from caproto import ChannelType
from caproto.server import PVGroup, SubGroup, ioc_arg_parser, pvproperty, run

from ophyd.tests.scaler_ioc import EpicsScalerGroup


class SpectroscopyScalerGroup(EpicsScalerGroup):
    offset_start = pvproperty(value=0, name="_offset_start.PROC")
    offset_time = pvproperty(value=5.0, name="_offset_time.VAL")

    class CalcsGroup(PVGroup):
        calc1 = pvproperty(value=2.35, name="_calc1.VAL", dtype=float)
        calc2 = pvproperty(value=2.35, name="_calc2.VAL", dtype=float)
        calc3 = pvproperty(value=2.35, name="_calc3.VAL", dtype=float)
        calc4 = pvproperty(value=2.35, name="_calc4.VAL", dtype=float)
        calc5 = pvproperty(value=2.35, name="_calc5.VAL", dtype=float)
        calc6 = pvproperty(value=2.35, name="_calc6.VAL", dtype=float)
        calc7 = pvproperty(value=2.35, name="_calc7.VAL", dtype=float)
        calc8 = pvproperty(value=2.35, name="_calc8.VAL", dtype=float)

    calcs = SubGroup(CalcsGroup, prefix="")

    class OffsetsGroup(PVGroup):
        offset1 = pvproperty(value=500000, name="_offset0.A", dtype=int)
        offset2 = pvproperty(value=500000, name="_offset0.B", dtype=int)
        offset3 = pvproperty(value=500000, name="_offset0.C", dtype=int)
        offset4 = pvproperty(value=500000, name="_offset0.D", dtype=int)
        offset5 = pvproperty(value=500000, name="_offset1.A", dtype=int)
        offset6 = pvproperty(value=500000, name="_offset1.B", dtype=int)
        offset7 = pvproperty(value=500000, name="_offset1.C", dtype=int)
        offset8 = pvproperty(value=500000, name="_offset1.D", dtype=int)
        offset9 = pvproperty(value=500000, name="_offset2.A", dtype=int)
        offset10 = pvproperty(value=500000, name="_offset2.B", dtype=int)
        offset11 = pvproperty(value=500000, name="_offset2.C", dtype=int)
        offset12 = pvproperty(value=500000, name="_offset2.D", dtype=int)
        offset13 = pvproperty(value=500000, name="_offset3.A", dtype=int)
        offset14 = pvproperty(value=500000, name="_offset3.B", dtype=int)
        offset15 = pvproperty(value=500000, name="_offset3.C", dtype=int)
        offset16 = pvproperty(value=500000, name="_offset3.D", dtype=int)
        offset17 = pvproperty(value=500000, name="_offset4.A", dtype=int)
        offset18 = pvproperty(value=500000, name="_offset4.B", dtype=int)
        offset19 = pvproperty(value=500000, name="_offset4.C", dtype=int)
        offset20 = pvproperty(value=500000, name="_offset4.D", dtype=int)
        offset21 = pvproperty(value=500000, name="_offset5.A", dtype=int)
        offset22 = pvproperty(value=500000, name="_offset5.B", dtype=int)
        offset23 = pvproperty(value=500000, name="_offset5.C", dtype=int)
        offset24 = pvproperty(value=500000, name="_offset5.D", dtype=int)
        offset25 = pvproperty(value=500000, name="_offset6.A", dtype=int)
        offset27 = pvproperty(value=500000, name="_offset6.B", dtype=int)
        offset27 = pvproperty(value=500000, name="_offset6.C", dtype=int)
        offset28 = pvproperty(value=500000, name="_offset6.D", dtype=int)
        offset29 = pvproperty(value=500000, name="_offset7.A", dtype=int)
        offset30 = pvproperty(value=500000, name="_offset7.B", dtype=int)
        offset31 = pvproperty(value=500000, name="_offset7.C", dtype=int)
        offset32 = pvproperty(value=500000, name="_offset7.D", dtype=int)

    offsets = SubGroup(OffsetsGroup, prefix="")

    class NetChannelsGroup(PVGroup):
        net1 = pvproperty(value=20000000, name="_netA.A", dtype=int)
        net2 = pvproperty(value=20000000, name="_netA.B", dtype=int)
        net3 = pvproperty(value=20000000, name="_netA.C", dtype=int)
        net4 = pvproperty(value=20000000, name="_netA.D", dtype=int)
        net5 = pvproperty(value=20000000, name="_netA.E", dtype=int)
        net6 = pvproperty(value=20000000, name="_netA.F", dtype=int)
        net7 = pvproperty(value=20000000, name="_netA.G", dtype=int)
        net8 = pvproperty(value=20000000, name="_netA.H", dtype=int)
        net9 = pvproperty(value=20000000, name="_netA.I", dtype=int)
        net10 = pvproperty(value=20000000, name="_netA.J", dtype=int)
        net11 = pvproperty(value=20000000, name="_netA.K", dtype=int)
        net12 = pvproperty(value=20000000, name="_netA.L", dtype=int)
        net13 = pvproperty(value=20000000, name="_netB.A", dtype=int)
        net14 = pvproperty(value=20000000, name="_netB.B", dtype=int)
        net15 = pvproperty(value=20000000, name="_netB.C", dtype=int)
        net16 = pvproperty(value=20000000, name="_netB.D", dtype=int)
        net17 = pvproperty(value=20000000, name="_netB.E", dtype=int)
        net18 = pvproperty(value=20000000, name="_netB.F", dtype=int)
        net19 = pvproperty(value=20000000, name="_netB.G", dtype=int)
        net20 = pvproperty(value=20000000, name="_netB.H", dtype=int)
        net21 = pvproperty(value=20000000, name="_netB.I", dtype=int)
        net22 = pvproperty(value=20000000, name="_netB.J", dtype=int)
        net23 = pvproperty(value=20000000, name="_netB.K", dtype=int)
        net24 = pvproperty(value=20000000, name="_netB.L", dtype=int)
        net25 = pvproperty(value=20000000, name="_netC.A", dtype=int)
        net26 = pvproperty(value=20000000, name="_netC.B", dtype=int)
        net27 = pvproperty(value=20000000, name="_netC.C", dtype=int)
        net28 = pvproperty(value=20000000, name="_netC.D", dtype=int)
        net29 = pvproperty(value=20000000, name="_netC.E", dtype=int)
        net30 = pvproperty(value=20000000, name="_netC.F", dtype=int)
        net31 = pvproperty(value=20000000, name="_netC.G", dtype=int)
        net32 = pvproperty(value=20000000, name="_netC.H", dtype=int)

    nets = SubGroup(NetChannelsGroup, prefix="")


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="255idVME:scaler1", desc="haven.tests.ioc_scaler test IOC"
    )
    ioc = SpectroscopyScalerGroup(**ioc_options)
    run(ioc.pvdb, **run_options)



import time
from ophyd import Kind

import numpy as np
from pyqtgraph import PlotItem
import pytest
from qtpy import QtCore

from firefly.main_window import FireflyMainWindow
from firefly.xrf_detector import XRFDetectorDisplay, XRFPlotWidget
from firefly.xrf_roi import XRFROIDisplay


@pytest.fixture()
def xrf_display(ffapp, sim_vortex):
    FireflyMainWindow()
    display = XRFDetectorDisplay(macros={"DEV": sim_vortex.name})
    spectra = np.random.default_rng(seed=0).integers(
        0, 65536, dtype=np.int_, size=(4, 1024)
    )
    plot_widget = display.mca_plot_widget
    plot_widget.update_spectrum(1, spectra[0])
    plot_widget.update_spectrum(2, spectra[1])
    return display


def test_open_xrf_detector_viewer_actions(ffapp, qtbot, sim_vortex):
    # Get the area detector parts ready
    ffapp.prepare_xrf_detector_windows()
    assert hasattr(ffapp, "xrf_detector_actions")
    assert len(ffapp.xrf_detector_actions) == 1
    # Launch an action and see that a window opens
    list(ffapp.xrf_detector_actions.values())[0].trigger()
    assert "FireflyMainWindow_xrf_detector_vortex_me4" in ffapp.windows.keys()


def test_roi_widgets(ffapp, sim_vortex):
    FireflyMainWindow()
    display = XRFDetectorDisplay(macros={"DEV": sim_vortex.name})
    display.draw_roi_widgets(2)
    # Check that the widgets were drawn
    assert len(display.roi_displays) == sim_vortex.num_rois
    disp = display.roi_displays[0]


def test_roi_element_comboboxes(ffapp, qtbot, sim_vortex):
    FireflyMainWindow()
    display = XRFDetectorDisplay(macros={"DEV": sim_vortex.name})
    # Check that the comboboxes have the right number of entries
    element_cb = display.ui.mca_combobox
    assert element_cb.count() == sim_vortex.num_elements
    roi_cb = display.ui.roi_combobox
    assert roi_cb.count() == sim_vortex.num_rois


def test_roi_selection(ffapp, qtbot, sim_vortex):
    FireflyMainWindow()
    display = XRFROIDisplay(
        macros={"DEV": sim_vortex.name, "NUM": 2, "MCA": 2, "ROI": 2}
    )
    # Unchecked box should be bland
    assert "background" not in display.styleSheet()
    # Make the ROI selected and check for a distinct background
    display.ui.set_roi_button.setChecked(True)
    assert f"background: {display.selected_background}" in display.styleSheet()
    # Disabling the ROI should unselect it as well
    display.ui.set_roi_button.setChecked(True)
    display.ui.enabled_checkbox.setChecked(True)
    display.ui.enabled_checkbox.setChecked(False)
    assert not display.ui.set_roi_button.isChecked()
    assert f"background: {display.selected_background}" not in display.styleSheet()


def test_all_rois_selection(ffapp, qtbot, sim_vortex):
    """Are all the other ROIs disabled when one is selected?"""
    FireflyMainWindow()
    display = XRFDetectorDisplay(macros={"DEV": sim_vortex.name})
    roi_display = display.roi_displays[0]
    # Pretend an ROI display was selected
    roi_display.selected.emit(True)
    # Check that a different ROI display was disabled
    assert not display.roi_displays[1].isEnabled()


def test_all_mcas_selection(ffapp, qtbot, sim_vortex):
    """Are all the other ROIs disabled when one is selected?"""
    FireflyMainWindow()
    display = XRFDetectorDisplay(macros={"DEV": sim_vortex.name})
    mca_display = display.mca_displays[0]
    # Pretend an ROI display was selected
    mca_display.selected.emit(True)
    # Check that a different ROI display was disabled
    assert not display.mca_displays[1].isEnabled()


def test_update_roi_spectra(ffapp, qtbot, sim_vortex):
    FireflyMainWindow()
    display = XRFDetectorDisplay(macros={"DEV": sim_vortex.name})
    spectra = np.random.default_rng(seed=0).integers(
        0, 65536, dtype=np.int_, size=(4, 1024)
    )
    roi_plot_widget = display.ui.roi_plot_widget
    with qtbot.waitSignal(roi_plot_widget.plot_changed):
        display._spectrum_channels[0].value_slot(spectra[0])
        display._spectrum_channels[1].value_slot(spectra[1])
    # Check that a PlotItem was created
    plot_item = roi_plot_widget.ui.plot_widget.getPlotItem()
    assert isinstance(plot_item, PlotItem)
    # Check that the spectrum was plotted
    data_items = plot_item.listDataItems()
    assert len(data_items) == 1
    # Check that previous plots get cleared
    spectra2 = np.random.default_rng(seed=1).integers(
        0, 65536, dtype=np.int_, size=(4, 1024)
    )
    with qtbot.waitSignal(roi_plot_widget.plot_changed):
        display._spectrum_channels[0].value_slot(spectra2[0])
    data_items = plot_item.listDataItems()
    assert len(data_items) == 1


def test_update_mca_spectra(ffapp, qtbot, sim_vortex):
    FireflyMainWindow()
    display = XRFDetectorDisplay(macros={"DEV": sim_vortex.name})
    spectra = np.random.default_rng(seed=0).integers(
        0, 65536, dtype=np.int_, size=(4, 1024)
    )
    mca_plot_widget = display.ui.mca_plot_widget
    # plot_widget.update_spectrum(spectrum=spectra[0], mca_idx=1)
    with qtbot.waitSignal(mca_plot_widget.plot_changed):
        display._spectrum_channels[0].value_slot(spectra[0])
        display._spectrum_channels[1].value_slot(spectra[1])
    # Check that a PlotItem was created
    plot_item = mca_plot_widget.ui.plot_widget.getPlotItem()
    assert isinstance(plot_item, PlotItem)
    # Check that the spectrum was plotted
    data_items = plot_item.listDataItems()
    assert len(data_items) == 2
    # Check that previous plots get cleared
    spectra2 = np.random.default_rng(seed=1).integers(
        0, 65536, dtype=np.int_, size=(4, 1024)
    )
    with qtbot.waitSignal(mca_plot_widget.plot_changed):
        display._spectrum_channels[0].value_slot(spectra2[0])
    data_items = plot_item.listDataItems()
    assert len(data_items) == 2


def test_mca_selected_highlights(ffapp, qtbot, sim_vortex):
    """Is the spectrum highlighted when the element row is selected."""
    FireflyMainWindow()
    display = XRFDetectorDisplay(macros={"DEV": sim_vortex.name})
    mca_display = display.mca_displays[1]
    spectra = np.random.default_rng(seed=0).integers(
        0, 65536, dtype=np.int_, size=(4, 1024)
    )
    plot_widget = display.mca_plot_widget
    plot_widget.update_spectrum(1, spectra[0])
    plot_widget.update_spectrum(2, spectra[1])
    this_data_item = display.mca_plot_widget._data_items[2]
    other_data_item = display.mca_plot_widget._data_items[1]
    this_data_item.setOpacity(0.77)
    # Select this display and check if the spectrum is highlighted
    mca_display.selected.emit(True)
    assert this_data_item.opacity() == 1.0
    assert other_data_item.opacity() == 0.15
    # Hovering other rows should not affect the opacity
    display.mca_displays[0].enterEvent()
    assert this_data_item.opacity() == 0.55
    assert other_data_item.opacity() == 1.0
    display.mca_displays[0].leaveEvent()
    assert this_data_item.opacity() == 1.0
    assert other_data_item.opacity() == 0.15


def test_show_mca_region_visibility(ffapp, xrf_display):
    """Is the spectrum highlighted when the element row is selected."""
    # Check that the region is hidden at startup
    plot_widget = xrf_display.mca_plot_widget
    region = plot_widget.region(mca_num=2, roi_num=0)
    assert not region.isVisible()
    # Now highlight a spectrum, and confirm it is visible
    plot_widget.highlight_spectrum(mca_num=2, roi_num=0, hovered=True)
    assert region.isVisible()
    # assert region.brush.color().name() == "#ff7f0e"
    # Unhighlight and confirm it is invisible
    plot_widget.highlight_spectrum(mca_num=1, roi_num=0, hovered=False)
    assert not region.isVisible()


def test_show_roi_region(ffapp, xrf_display):
    """Is the spectrum highlighted when the element row is selected."""
    # Check that the region is hidden at startup
    plot_widget = xrf_display.roi_plot_widget
    hovered_region = plot_widget.region(mca_num=1, roi_num=1)
    plot_widget.select_roi(mca_num=1, roi_num=0, is_selected=True)
    selected_region = plot_widget.region(mca_num=1, roi_num=0)
    assert not hovered_region.isVisible()
    assert selected_region.isVisible()
    # Now highlight a spectrum, and confirm it is visible
    plot_widget.highlight_spectrum(mca_num=1, roi_num=1, hovered=True)
    assert hovered_region.isVisible()
    assert not selected_region.isVisible()
    # assert region.brush.color().name() == "#ff7f0e"
    # Unhighlight and confirm it is invisible
    plot_widget.highlight_spectrum(mca_num=1, roi_num=0, hovered=False)
    assert not hovered_region.isVisible()
    assert selected_region.isVisible()


def test_mca_region_channels(ffapp, xrf_display):
    """Are the channel access connections between the ROI selection region
    and the hi/lo channel PVs correct?

    """
    plot_widget = xrf_display.mca_plot_widget
    plot_widget.device_name = "vortex_me4"
    mca_display = xrf_display.mca_displays[1]
    mca_display._embedded_widget = mca_display.open_file(force=True)
    xrf_display.mca_selected(is_selected=True, mca_num=2)
    correct_address = "oph://vortex_me4.mcas.mca2.rois.roi0.hi_chan._write_pv"
    region = plot_widget.region(mca_num=2, roi_num=0)
    assert region.hi_channel.address == correct_address
    region.hi_channel.value_slot(108)
    assert region.getRegion()[1] == 108
    region.lo_channel.value_slot(47)
    assert region.getRegion()[0] == 47


def test_mca_copyall_button(ffapp, xrf_display, qtbot):
    xrf_display.mca_selected(is_selected=True, mca_num=2)
    assert xrf_display.ui.mca_copyall_button.isEnabled()
    # Set up ROI displays to test
    this_display = xrf_display.mca_displays[1]
    this_display._embedded_widget = this_display.open_file(force=True)
    other_display = xrf_display.mca_displays[0]
    other_display._embedded_widget = other_display.open_file(force=True)
    # Change the values on the MCA displays
    this_display.embedded_widget.ui.lower_lineedit.setText("111")
    this_display.embedded_widget.ui.upper_lineedit.setText("131")
    this_display.embedded_widget.ui.label_lineedit.setText("Ni Ka")
    # Copy to the other MCA display
    qtbot.mouseClick(xrf_display.ui.mca_copyall_button, QtCore.Qt.LeftButton)
    assert other_display.embedded_widget.ui.lower_lineedit.text() == "111"
    assert other_display.embedded_widget.ui.upper_lineedit.text() == "131"
    # Does the button get disabled on un-select?
    xrf_display.mca_selected(is_selected=False, mca_num=2)
    assert not xrf_display.ui.mca_copyall_button.isEnabled()


def test_roi_copyall_button(ffapp, xrf_display, qtbot):
    # Set up ROI rows embedded display widgets
    for disp in xrf_display.roi_displays:
        disp._embedded_widget = disp.open_file(force=True)
    # Select an ROI
    xrf_display.roi_selected(is_selected=True, roi_num=1)
    assert xrf_display.ui.roi_copyall_button.isEnabled()
    # Set up ROI displays to test
    this_display = xrf_display.roi_displays[1]
    this_display._embedded_widget = this_display.open_file(force=True)
    other_display = xrf_display.roi_displays[0]
    other_display._embedded_widget = other_display.open_file(force=True)
    # Change the values on the MCA displays
    this_display.embedded_widget.ui.lower_lineedit.setText("111")
    this_display.embedded_widget.ui.upper_lineedit.setText("131")
    this_display.embedded_widget.ui.label_lineedit.setText("Ni Ka")
    # Copy to the other ROI display
    qtbot.mouseClick(xrf_display.ui.roi_copyall_button, QtCore.Qt.LeftButton)
    assert other_display.embedded_widget.ui.lower_lineedit.text() == "111"
    assert other_display.embedded_widget.ui.upper_lineedit.text() == "131"
    # Does the button get disabled on un-select?
    xrf_display.roi_selected(is_selected=False, roi_num=1)
    assert not xrf_display.ui.roi_copyall_button.isEnabled()


def test_mca_enableall_checkbox(ffapp, xrf_display):
    checkbox = xrf_display.ui.mca_enableall_checkbox
    assert checkbox.checkState() == QtCore.Qt.PartiallyChecked
    assert checkbox.isTristate()
    for display in xrf_display.mca_displays:
        display._embedded_widget = display.open_file(force=True)
        assert not display.embedded_widget.ui.enabled_checkbox.checkState()
    # Set it to checked and make sure all the ROI checkboxes respond
    checkbox.setCheckState(QtCore.Qt.Checked)
    assert not checkbox.isTristate()
    for display in xrf_display.mca_displays:
        assert display.embedded_widget.ui.enabled_checkbox.isChecked()
    # Un-enable all, does it go back?
    checkbox.setCheckState(QtCore.Qt.Unchecked)
    for display in xrf_display.mca_displays:
        assert not display.embedded_widget.ui.enabled_checkbox.isChecked()


def test_roi_enableall_checkbox(ffapp, xrf_display):
    checkbox = xrf_display.ui.roi_enableall_checkbox
    assert checkbox.checkState() == QtCore.Qt.PartiallyChecked
    assert checkbox.isTristate()
    for display in xrf_display.roi_displays:
        display._embedded_widget = display.open_file(force=True)
        assert not display.embedded_widget.ui.enabled_checkbox.checkState()
    # Set it to checked and make sure all the ROI checkboxes respond
    checkbox.setCheckState(QtCore.Qt.Checked)
    assert not checkbox.isTristate()
    for display in xrf_display.roi_displays:
        assert display.embedded_widget.ui.enabled_checkbox.isChecked()
    # Un-enable all, does it go back?
    checkbox.setCheckState(QtCore.Qt.Unchecked)
    for display in xrf_display.roi_displays:
        assert not display.embedded_widget.ui.enabled_checkbox.isChecked()


def test_oneshot_acquisition(xrf_display, qtbot):
    """Check that clicking the one-shot acquisition button works."""
    with qtbot.waitSignal(xrf_display.triggers.start_erase) as val:
        xrf_display.ui.oneshot_button.click()
    # Check the accumulate box and see that we don't erase
    xrf_display.ui.accumulate_checkbox.setChecked(True)
    with qtbot.waitSignal(xrf_display.triggers.start_all) as val:
        xrf_display.ui.oneshot_button.click()


def test_continuous_acquisition(xrf_display, qtbot):
    """Check that clicking the one-shot acquisition button works."""
    with qtbot.waitSignal(xrf_display.triggers.start_erase) as val:
        xrf_display.ui.continuous_button.click()
    # Simulated acquisition finishing and then set up the next one
    with qtbot.waitSignal(xrf_display.triggers.start_erase) as val:
        xrf_display.triggers.acquiring_channel.value_slot(0)



import time
import pytest

from firefly.main_window import FireflyMainWindow, PlanMainWindow
from firefly.application import FireflyApplication


def test_navbar(ffapp):
    window = PlanMainWindow()
    # Check navbar actions on the app
    assert hasattr(ffapp, "pause_runengine_action")
    # Check that the navbar actions are set up properly
    assert hasattr(window.ui, "navbar")
    navbar = window.ui.navbar
    # Navigation actions are removed
    assert window.ui.actionHome not in navbar.actions()
    # Run engine actions have been added to the navbar
    assert ffapp.pause_runengine_action in navbar.actions()
    assert ffapp.start_queue_action in navbar.actions()


def test_navbar_autohide(ffapp, qtbot):
    """Test that the queue navbar is only visible when plans are queued."""
    window = PlanMainWindow()
    window.show()
    navbar = window.ui.navbar
    # Pretend the queue has some things in it
    with qtbot.waitSignal(ffapp.queue_length_changed):
        ffapp.queue_length_changed.emit(3)
    assert navbar.isVisible()
    # Make the queue be empty
    with qtbot.waitSignal(ffapp.queue_length_changed):
        ffapp.queue_length_changed.emit(0)
    assert not navbar.isVisible()


def test_add_menu_action(ffapp):
    window = FireflyMainWindow()
    # Check that it's not set up with the right menu yet
    assert not hasattr(window, "actionMake_Salad")
    # Add a menu item
    action = window.add_menu_action(
        action_name="actionMake_Salad", text="Make Salad", menu=window.ui.menuTools
    )
    assert hasattr(window.ui, "actionMake_Salad")
    assert hasattr(window, "actionMake_Salad")
    assert action.text() == "Make Salad"
    assert action.objectName() == "actionMake_Salad"


def test_customize_ui(ffapp):
    window = FireflyMainWindow()
    assert hasattr(window.ui, "menuScans")


def test_show_message(ffapp):
    window = FireflyMainWindow()
    status_bar = window.statusBar()
    # Send a message
    window.show_status("Hello, APS.")



import unittest
import logging

import numpy as np

from haven import ERange, KRange, merge_ranges


logging.basicConfig(level=logging.INFO)


class EnergyRangeTests(unittest.TestCase):
    def test_e_range(self):
        """Test the ERange class for calculating energy points."""
        e_range = ERange(E_min=8300, E_max=8400, E_step=0.5, exposure=0.1)
        np.testing.assert_equal(e_range.energies(), np.arange(8300, 8400.5, 0.5))
        np.testing.assert_equal(e_range.exposures(), [0.1] * len(e_range.energies()))

    def test_k_range(self):
        E0 = 17038
        k_range = KRange(E_min=30, k_max=14, k_step=0.05, k_weight=1, exposure=1.0)
        # Verify the results compared to Shelly's spreadsheet
        np.testing.assert_almost_equal(k_range.wavenumbers()[0], 2.8060742521)
        np.testing.assert_almost_equal(k_range.wavenumbers()[-1], 14.006074252)
        energies = k_range.energies() + E0
        np.testing.assert_equal(energies[0], E0 + k_range.E_min)
        np.testing.assert_almost_equal(energies[-1], 17785.40463351)
        np.testing.assert_equal(k_range.exposures()[0], 1.0)
        np.testing.assert_almost_equal(
            k_range.exposures()[16], 1.28529317348436, decimal=3
        )

    def test_merge_ranges(self):
        e_range1 = ERange(1, 5, 1, exposure=0.5)
        e_range2 = ERange(5, 7, 0.5, exposure=1)
        merged, exposures = merge_ranges(e_range2, e_range1)
        # Test validity of the result
        np.testing.assert_equal(merged, [1, 2, 3, 4, 5, 5.5, 6, 6.5, 7])
        np.testing.assert_equal(
            exposures, [0.5, 0.5, 0.5, 0.5, 1.0, 1.0, 1.0, 1.0, 1.0]
        )



from apstools.synApps.db_2slit import Optics2Slit2D_HV

from .._iconfig import load_config
from .instrument_registry import registry


def load_slits(config=None):
    if config is None:
        config = load_config()
    # Create slits
    for name, slit_config in config.get("slits", {}).items():
        dev = Optics2Slit2D_HV(
            prefix=slit_config["prefix"], name=name, labels={"slits"}
        )
        registry.register(dev)



import logging
import warnings
import unicodedata
import re
import datetime as dt
from pathlib import Path
from typing import Sequence, Optional, Union

from bluesky.callbacks import CallbackBase

from . import exceptions


log = logging.getLogger(__name__)


def slugify(value, allow_unicode=False):
    """
    Taken from https://github.com/django/django/blob/master/django/utils/text.py
    Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
    dashes to single dashes. Remove characters that aren't alphanumerics,
    underscores, or hyphens. Also strip leading and
    trailing whitespace, dashes, and underscores.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = (
            unicodedata.normalize("NFKD", value)
            .encode("ascii", "ignore")
            .decode("ascii")
        )
    return re.sub(r"[-\s]+", "-", value).strip("-_")


class XDIWriter(CallbackBase):
    """A callback class for bluesky that will save data to an XDI file.

    The XAFS data interchange (XDI) format is a plain text,
    tab-separated file format that standardizes XAFS data. Metadata is
    also included in headers.

    The location of the file is determined by passing the *filename*
    parameter in when created the writer object:

    .. code-block:: python

      writer = XDIWriter(fd="./my_experiment.xdi")

    Placeholders can be included that will be filled-in with metadata
    during the start phase of the plan, for example:

    .. code-block:: python

      plan = energy_scan(..., E0="Ni_K", md=dict(sample_name="nickel oxide"))
      writer_callback = XDIWriter(fd="./{year}{month}{day}_{sample_name}_{edge}.xdi")
      RE(plan, writer)

    Assuming the date is 2022-08-19, then the filename will become
    "20220819_nickel-oxide_Ni_K.xdi".

    *{year}*, *{month}*, *{day}* describe the numerical, zero-padded
    year, month and day when the callback handles the start document
    for the plan. The remaining fields in the start document are also
    available, and will likely vary from plan to plan. Either consult
    the documentation for the plan being executed, or inspect the logs
    for the dictionary metadata at level logging.INFO when using an
    invalid placeholder.

    Parameters
    ==========
    fd
      Either an open file with write intent, or a Path or string with
      the file location. If *fd* is not already an open file, it will
      be created when the ``start()`` method is called.

    """

    _fd = None
    fp: Optional[Union[str, Path]] = None
    column_names: Optional[Sequence[str]] = None
    start_time: dt.datetime = None

    def __init__(self, fd, *args, **kwargs):
        is_file_obj = hasattr(fd, "writable")
        if is_file_obj:
            # *fd* is an open file object
            # Make sure it is writable
            if fd.writable():
                log.debug(f"Found open, writable file: {fd}")
                self._fd = fd
            else:
                msg = f"No write intent on file: {fd}"
                log.error(msg)
                raise exceptions.FileNotWritable(msg)
        else:
            # Assume *fd* is a path to a file
            self.fp = Path(fd).expanduser()
        return super().__init__(*args, **kwargs)

    @property
    def fd(self):
        """Retrieve the open writable file object for this writer."""
        if self._fd is None:
            log.debug(f"Opening file: {self.fp.resolve()}")
            self._fd = open(self.fp, mode="x")
        return self._fd

    def _path_metadata(self, doc={}):
        """Prepare the metadata for string formatting the output file path."""
        md = {
            "year": self.start_time.strftime("%Y"),
            "month": self.start_time.strftime("%m"),
            "day": self.start_time.strftime("%d"),
        }
        md.update(doc)
        return md

    def start(self, doc):
        self.start_time = dt.datetime.now().astimezone()
        # Format the file name based on metadata
        if self.fp is not None:
            fp = str(self.fp)
            md = self._path_metadata(doc=doc)
            try:
                fp = fp.format(**md)
            except KeyError as e:
                msg = f"Could not find match key {e} in {fp}."
                log.error(msg)
                log.info(f"Metadata is: {md}")
                raise exceptions.XDIFilenameKeyNotFound(msg) from None
            fp = slugify(fp)
            self.fp = Path(fp)
        # Save the rest of the start doc so we can write the headers
        # when we get our first datum
        self.start_doc = doc
        # Open the file, just to be sure we can
        self.fd

    def stop(self, doc):
        self.fd.close()

    def write_header(self, doc):
        fd = self.fd
        # Write package version information
        versions = ["XDI/1.0"]
        versions += [f"{name}/{ver}" for name, ver in doc.get("versions", {}).items()]
        fd.write(f"# {' '.join(versions)}\n")
        # Column Names
        columns = [
            f"# Column.{num+1}: {name}\n" for num, name in enumerate(self.column_names)
        ]
        fd.write("".join(columns))
        # X-ray edge information
        edge_str = doc.get("edge", None)
        try:
            elem, edge = edge_str.split("_")
        except (AttributeError, ValueError):
            msg = f"Could not parse X-ray edge metadata: {edge_str}"
            warnings.warn(msg)
            log.warning(msg)
        else:
            fd.write(f"# Element.symbol: {elem}\n")
            fd.write(f"# Element.edge: {edge}\n")
        now = self.start_time
        fd.write(f"# Scan.start_time: {now.strftime('%Y-%m-%d %H:%M:%S%z')}\n")
        md_paths = [
            "facility.name",
            "facility.xray_source",
            "beamline.name",
            "beamline.pv_prefix",
        ]
        for path in md_paths:
            val = doc
            try:
                for piece in path.split("."):
                    val = val[piece]
                fd.write(f"# {path}: {val}\n")
            except KeyError:
                continue
        # Scan ID
        fd.write(f"# uid: {self.start_doc.get('uid', '')}\n")
        # Header end token
        fd.write("# -------------\n")

    def event(self, doc):
        """Save the data in tab-separated value, in order of
        ``self.column_names``.

        """
        data = doc["data"]
        # Use metadata from the first event to finish writing the header
        if self.column_names is None:
            names = list(data.keys())
            # Sort column names so that energy-related fields are first
            names = sorted(names, key=lambda x: not x.startswith("energy"))
            self.column_names = names + ["time"]
            self.write_header(self.start_doc)
        # Read in and store the actual data
        fd = self.fd
        values = []
        for col in self.column_names:
            if col == "time":
                values.append(str(doc["time"]))
            else:
                values.append(str(data[col]))
        line = "\t".join(values)
        fd.write(line + "\n")



from typing import Union, Sequence

from bluesky import plan_stubs as bps
from ophyd import Device

from ..instrument.instrument_registry import registry
from ..typing import DetectorList
from .. import exceptions


__all__ = ["set_energy"]


Harmonic = Union[str, int, None]


def auto_harmonic(energy: float, harmonic: Harmonic) -> Harmonic:
    # No harmonic change is requested
    if harmonic is None:
        return harmonic
    # Check for specific harmonics
    try:
        return int(harmonic)
    except ValueError:
        pass
    # Check for auto harmonic selection
    threshold = 11000
    if harmonic == "auto":
        if energy < threshold:
            return 1
        else:
            return 3
    # If we get here, the harmonic was not a valid option
    raise exceptions.InvalidHarmonic(
        f"Insertion device cannot accept harmonic: {harmonic}"
    )


def set_energy(
    energy: float,
    harmonic: Harmonic = None,
    positioners: DetectorList = ["energy"],
    harmonic_positioners: DetectorList = ["undulator_harmonic_value"],
):
    """Set the energy of the beamline, in electron volts.

    Moves both the mono energy, and the undulator energy with a
    calibrated offset.

    The *harmonic* parameter selects a harmonic for the undulator. If
    ``"auto"``, then the harmonic will be selected based on the
    energy. If *harmonic* is ``None``, then the current harmonic is
    used. If *harmonic* is an integer (e.g. 1, 3, 5) then this value
    will be used.

    Parameters
    ==========
    energy
      The target energy of the beamline, in electron-volts.
    harmonic

      Which harmonic to use for the undulator. Can be an integer
      (e.g. 1, 3, 5), ``None``, or ``"auto"``.

    """
    # Prepare arguments for undulator harmonic
    harmonic = auto_harmonic(energy, harmonic)
    if harmonic is not None:
        harmonic_positioners = registry.findall(name=harmonic_positioners)
        hargs = []
        for positioner in harmonic_positioners:
            hargs.extend([positioner, harmonic])
        yield from bps.mv(*hargs)
    # Prepare arguments for energy
    positioners = registry.findall(name=positioners)
    args = []
    for positioner in positioners:
        args.extend([positioner, energy])
    # Execute the plan
    yield from bps.mv(*args)



import logging

from haven import registry, exceptions
from pydm.data_plugins import add_plugin
from pydm.data_plugins.epics_plugins.pyepics_plugin_component import (
    Connection,
    PyEPICSPlugin,
)


log = logging.getLogger(__name__)


class OphydConnection(Connection):
    """A pydm connection for retrieving ophyd signals from haven registry.

    If the signal has a valid PV, then this connection functions as a
    conventional PyDMEpicsConnection. The *name* argument should match
    the *name* given when constructing the ophyd signal. The signal
    should be a leaf of the device hierarchy, i.e. one that
    corresponds to an EPICS PV.

    If the signal is not found, or does not have a PV, then the name
    will be used as is so that widgets display suitable feedback.

    """

    def __init__(self, channel, name, protocol=None, parent=None):
        # Resolve the device based on the ohpyd name
        try:
            component = registry.find(name)
        except (exceptions.ComponentNotFound, AttributeError):
            # Component does not exist, so make a dummy component
            log.warning(f"Could not find component {name} in instrument registry.")
            component = type("NullDevice", (), {"pvname": name})
        try:
            pv = component.pvname
        except AttributeError:
            pv = name
            log.warning(
                f"Component {name} does not have a pv, consider using one of its children."
            )
        # Use the new PV to get a regular PV connection
        log.debug(f"Converted ophyd name {name} to PV {pv}")
        super().__init__(channel=channel, pv=pv, protocol=protocol, parent=parent)


class OphydPlugin(PyEPICSPlugin):
    protocol = "oph"
    connection_class = OphydConnection



import pytest
from unittest import TestCase, expectedFailure
from io import StringIO
import os
import time
from pathlib import Path
import datetime as dt
import logging

# from freezegun import freeze_time
import pytz
import time_machine
from bluesky import RunEngine
from ophyd.sim import motor, SynGauss, SynAxis
import numpy as np
from numpy import asarray as array


from haven import XDIWriter, exceptions, energy_scan


fake_time = pytz.timezone("America/New_York").localize(
    dt.datetime(2022, 8, 19, 19, 10, 51)
)


log = logging.getLogger(__name__)
# logging.basicConfig(level=logging.DEBUG)


# Stub the epics signals in Haven

THIS_DIR = Path(__file__).parent


# Sample metadata dict
{
    "E0": 0,
    "beamline": {
        "is_connected": False,
        "name": "SPC Beamline (sector unknown)",
        "pv_prefix": "",
    },
    "detectors": ["I0", "It"],
    "edge": "Ni_K",
    "facility": {"name": "Advanced Photon Source", "xray_source": "insertion device"},
    "hints": {"dimensions": [(["energy", "exposure"], "primary")]},
    "ion_chambers": {"scaler": {"pv_prefix": ""}},
    "motors": ["energy", "exposure"],
    "num_intervals": 9,
    "num_points": 10,
    "plan_args": {
        "args": [
            "SynAxis(prefix='', name='energy', "
            "read_attrs=['readback', 'setpoint'], "
            "configuration_attrs=['velocity', 'acceleration'])",
            array([8300, 8310, 8320, 8330, 8340, 8350, 8360, 8370, 8380, 8390]),
            "SynAxis(prefix='', name='exposure', "
            "read_attrs=['readback', 'setpoint'], "
            "configuration_attrs=['velocity', 'acceleration'])",
            [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1],
        ],
        "detectors": [
            "SynGauss(prefix='', name='I0', "
            "read_attrs=['val'], configuration_attrs=['Imax', "
            "'center', 'sigma', 'noise', 'noise_multiplier'])",
            "SynGauss(prefix='', name='It', "
            "read_attrs=['val'], configuration_attrs=['Imax', "
            "'center', 'sigma', 'noise', "
            "'noise_multiplier'])",
        ],
        "per_step": "None",
    },
    "plan_name": "list_scan",
    "plan_pattern": "inner_list_product",
    "plan_pattern_args": {
        "args": [
            "SynAxis(prefix='', name='energy', "
            "read_attrs=['readback', 'setpoint'], "
            "configuration_attrs=['velocity', "
            "'acceleration'])",
            array([8300, 8310, 8320, 8330, 8340, 8350, 8360, 8370, 8380, 8390]),
            "SynAxis(prefix='', name='exposure', "
            "read_attrs=['readback', 'setpoint'], "
            "configuration_attrs=['velocity', "
            "'acceleration'])",
            [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1],
        ]
    },
    "plan_pattern_module": "bluesky.plan_patterns",
    "plan_type": "generator",
    "sample_name": "NiO_rock_salt",
    "scan_id": 1,
    "time": 1661025010.409619,
    "uid": "671c3c48-f014-421d-b3e0-57991b6745f6",
    "versions": {"bluesky": "1.8.3", "ophyd": "1.6.4"},
}


@pytest.fixture()
def stringio():
    yield StringIO()


class CallbackTests(TestCase):
    stringio = None
    file_path = THIS_DIR / "sample_file.txt"
    start_doc = {
        "versions": {"bluesky": "1.8.3", "ophyd": "1.6.4"},
        "detectors": ["I0", "It"],
        "motors": ["energy", "exposure"],
        "edge": "Ni_K",
        "facility": {
            "name": "Advanced Photon Source",
            "xray_source": "insertion device",
        },
        "beamline": {"name": "20-ID-C", "pv_prefix": "20id:"},
        "sample_name": "nickel oxide",
        "uid": "671c3c48-f014-421d-b3e0-57991b6745f6",
    }
    event_doc = {
        "data": {"I0": 2, "It": 1.5, "energy": 8330, "exposure": 0.1},
        "time": 1660186828.0055554,
    }

    def setUp(self):
        self.stringio = StringIO()

    def tearDown(self):
        for fp in [
            self.file_path,
            Path("{year}{month}{day}_{sample_name}.xdi"),
            Path("20220819_nickel-oxide.xdi"),
        ]:
            if fp.exists():
                fp.unlink()

    def test_opens_file(self):
        # Pass in a string and it should hold the path until the start document is found
        writer = XDIWriter(self.file_path)
        self.assertEqual(writer.fp, self.file_path)
        self.assertFalse(self.file_path.exists())
        # Run the writer
        writer("start", {"edge": "Ni_K"})
        # Check that a file was created
        self.assertTrue(self.file_path.exists())

    def test_uses_open_file(self):
        # Pass in a string and it should hold the path until the start document is found
        writer = XDIWriter(self.stringio)
        self.assertIs(writer.fd, self.stringio)

    def test_read_only_file(self):
        # Check that it raises an exception if an open file has no write intent
        # Put some content in the temporary file
        with open(self.file_path, mode="w") as fd:
            fd.write("Hello, spam!")
        # Check that the writer raises an exception when the file is open read-only
        with open(self.file_path, mode="r") as fd:
            with self.assertRaises(exceptions.FileNotWritable):
                XDIWriter(fd)

    @expectedFailure
    def test_required_headers(self):
        writer = XDIWriter(self.stringio)
        writer("start", self.start_doc)
        # Check that required headers were added to the XDI file
        self.stringio.seek(0)
        xdi_output = self.stringio.read()
        self.assertIn("# XDI/1.0 bluesky/1.8.3 ophyd/1.6.4", xdi_output)
        self.assertIn("# Column.1: energy", xdi_output)
        self.assertIn("# Element.symbol: Ni", xdi_output)
        self.assertIn("# Element.edge: K", xdi_output)
        self.assertIn("# Mono.d_spacing: 3", xdi_output)
        self.assertIn("# -------------")

    @time_machine.travel(fake_time)
    def test_optional_headers(self):
        writer = XDIWriter(self.stringio)
        os.environ["TZ"] = "America/New_York"
        time.tzset()
        writer("start", self.start_doc)
        writer("event", self.event_doc)
        # Check that required headers were added to the XDI file
        self.stringio.seek(0)
        xdi_output = self.stringio.read()
        expected_metadata = {
            "Facility.name": "Advanced Photon Source",
            "Facility.xray_source": "insertion device",
            "Beamline.name": "20-ID-C",
            "Beamline.pv_prefix": "20id:",
            "Scan.start_time": "2022-08-19 19:10:51-0400",
            "Column.5": "time",
            "uid": "671c3c48-f014-421d-b3e0-57991b6745f6",
        }
        for key, val in expected_metadata.items():
            self.assertIn(f"# {key.lower()}: {val.lower()}\n", xdi_output.lower())

    @time_machine.travel(fake_time)
    def test_file_path_formatting(self):
        """Check that "{date}_{user}.xdi" formatting works in the filename."""
        writer = XDIWriter("{year}{month}{day}_{sample_name}.xdi")
        writer.start(self.start_doc)
        self.assertEqual(str(writer.fp), "20220819_nickel-oxide.xdi")

    def test_file_path_formatting_bad_key(self):
        """Check that "{date}_{user}.xdi" raises exception if placeholder is invalid."""
        writer = XDIWriter("{year}{month}{day}_{spam}_{sample_name}.xdi")
        with self.assertRaises(exceptions.XDIFilenameKeyNotFound):
            writer.start(self.start_doc)

    def test_data(self):
        """Check that the TSV data section is present and correct."""
        writer = XDIWriter(self.stringio)
        writer.start(self.start_doc)
        writer("event", self.event_doc)
        # Verify the data were written properly
        self.stringio.seek(0)
        xdi_output = self.stringio.read()
        self.assertIn("8330", xdi_output)
        self.assertIn("1660186828.0055554", xdi_output)


def test_with_plan(stringio, sim_registry):
    I0 = SynGauss(
        "I0",
        motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"detectors"},
    )
    It = SynGauss(
        "It",
        motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"detectors"},
    )
    energy_motor = SynAxis(name="energy", labels={"motors", "energies"})
    exposure = SynAxis(name="exposure", labels={"motors", "exposures"})
    writer = XDIWriter(stringio)
    RE = RunEngine()
    energy_plan = energy_scan(
        np.arange(8300.0, 8400.0, 10),
        detectors=[I0, It],
        E0="Ni_K",
        energy_positioners=[energy_motor],
        time_positioners=[exposure],
        md=dict(sample_name="NiO_rock_salt"),
    )
    RE(energy_plan, writer)



import json
import logging

from pydm.widgets import PyDMEmbeddedDisplay
import haven

from firefly import display


log = logging.getLogger(__name__)


class CamerasDisplay(display.FireflyDisplay):
    _camera_displays = []

    def __init__(self, args=None, macros={}, **kwargs):
        self._camera_displays = []
        super().__init__(args=args, macros=macros, **kwargs)

    def customize_ui(self):
        # Delete existing camera widgets
        for idx in reversed(range(self.cameras_layout.count())):
            self.cameras_layout.takeAt(idx).widget().deleteLater()
        # Add embedded displays for all the cameras
        try:
            cameras = haven.registry.findall(label="cameras")
        except haven.exceptions.ComponentNotFound:
            log.warning(
                "No cameras found, [Detectors] -> [Cameras] panel will be empty."
            )
            cameras = []
        for camera in sorted(cameras, key=lambda c: c.name):
            disp = PyDMEmbeddedDisplay(parent=self)
            disp.macros = json.dumps(
                {
                    "CAMERA": camera.name,
                    "DESC": camera.description,
                }
            )
            disp.filename = "camera.py"
            # Add the Embedded Display to the Results Layout
            self.cameras_layout.addWidget(disp)
            self._camera_displays.append(disp)

    def ui_filename(self):
        return "cameras.ui"



import datetime as dt
import logging
from collections import OrderedDict
from typing import Sequence

from qtpy.QtCore import QObject, Slot, Signal
from tiled import queries

from haven import tiled_client


log = logging.getLogger(__name__)


class DatabaseWorker(QObject):
    selected_runs: Sequence = []
    _filters = {"exit_status": "success"}

    # Signals
    all_runs_changed = Signal(list)
    selected_runs_changed = Signal(list)
    distinct_fields_changed = Signal(dict)
    new_message = Signal(str, int)
    db_op_started = Signal()
    db_op_ended = Signal(list)  # (list of exceptions thrown)

    def __init__(self, root_node, *args, **kwargs):
        if root_node is None:
            root_node = tiled_client()
        self.root = root_node
        super().__init__(*args, **kwargs)

    def set_filters(self, filters):
        log.debug(f"Setting new filters: {filters}")
        self._filters = filters

    def filtered_nodes(self):
        case_sensitive = False
        runs = self.root
        filters = self._filters
        log.debug(f"Filtering nodes: {filters}")
        filter_params = [
            # (filter_name, query type, metadata key)
            ("user", queries.Regex, "proposal_users"),
            ("proposal", queries.Regex, "proposal_id"),
            ("esaf", queries.Regex, "esaf_id"),
            ("sample", queries.Regex, "sample_name"),
            # ('exit_status', queries.Regex, "exit_status"),
            ("plan", queries.Regex, "plan_name"),
            ("edge", queries.Regex, "edge"),
        ]
        for filter_name, Query, md_name in filter_params:
            val = filters.get(filter_name, "")
            if val != "":
                runs = runs.search(Query(md_name, val, case_sensitive=case_sensitive))
        full_text = filters.get("full_text", "")
        if full_text != "":
            runs = runs.search(
                queries.FullText(full_text, case_sensitive=case_sensitive)
            )
        return runs

    @Slot()
    def load_distinct_fields(self):
        """Get distinct metadata fields for filterable metadata.

        Emits
        =====
        distinct_fields_changed
          Emitted with the new dictionary of distinct metadata choices
          for each metadata key.

        """
        new_fields = {}
        target_fields = [
            "sample_name",
            "proposal_users",
            "proposal_id",
            "esaf_id",
            "sample_name",
            "plan_name",
            "edge",
        ]
        # Get fields from the database
        response = self.root.distinct(*target_fields)
        # Build into a new dictionary
        for key, result in response["metadata"].items():
            field = key.split(".")[-1]
            new_fields[field] = [r["value"] for r in result]
        self.distinct_fields_changed.emit(new_fields)

    @Slot()
    def load_all_runs(self):
        all_runs = []
        nodes = self.filtered_nodes()
        self.db_op_started.emit()
        try:
            for uid, node in nodes.items():
                # Get meta-data documents
                metadata = node.metadata
                start_doc = metadata.get("start")
                if start_doc is None:
                    log.debug(f"Skipping run with no start doc: {uid}")
                    continue
                stop_doc = node.metadata.get("stop")
                if stop_doc is None:
                    stop_doc = {}
                # Get a human-readable timestamp for the run
                timestamp = start_doc.get("time")
                if timestamp is None:
                    run_datetime = ""
                else:
                    run_datetime = dt.datetime.fromtimestamp(timestamp)
                    run_datetime = run_datetime.strftime("%Y-%m-%d %H:%M:%S")
                # Get the X-ray edge scanned
                edge = start_doc.get("edge")
                E0 = start_doc.get("E0")
                E0_str = "" if E0 is None else str(E0)
                if edge and E0:
                    edge_str = f"{edge} ({E0} eV)"
                elif edge:
                    edge_str = edge
                elif E0:
                    edge_str = E0_str
                else:
                    edge_str = ""
                # Build the table item
                # Get sample data from: dd80f432-c849-4749-a8f3-bdeec6f9c1f0
                run_data = OrderedDict(
                    plan_name=start_doc.get("plan_name", ""),
                    sample_name=start_doc.get("sample_name", ""),
                    edge=edge_str,
                    E0=E0_str,
                    exit_status=stop_doc.get("exit_status", ""),
                    run_datetime=run_datetime,
                    uid=uid,
                    proposal_id=start_doc.get("proposal_id", ""),
                    esaf_id=start_doc.get("esaf_id", ""),
                    esaf_users=start_doc.get("esaf_users", ""),
                )
                all_runs.append(run_data)
        except Exception as exc:
            self.db_op_ended.emit([exc])
            raise
        else:
            self.db_op_ended.emit([])

        self.all_runs_changed.emit(all_runs)

    @Slot(list)
    def load_selected_runs(self, uids):
        # Retrieve runs from the database
        uids = list(set(uids))
        self.db_op_started.emit()
        # Download each item, maybe we can find a more efficient way to do this
        try:
            runs = [self.root[uid] for uid in uids]
            for run in runs:
                run["primary"].download()
        except Exception as exc:
            self.db_op_ended.emit([exc])
            raise
        else:
            self.db_op_ended.emit([])
        # Save and inform clients of the run data
        self.selected_runs = runs
        self.selected_runs_changed.emit(runs)



#!/usr/bin/env python3
from caproto.server import (
    PVGroup,
    SubGroup,
    ioc_arg_parser,
    pvproperty,
    run,
    PvpropertyInteger,
)

from ophyd.tests.scaler_ioc import EpicsScalerGroup


class AreaDetectorGroup(PVGroup):
    """An IOC matching an EPICS area detector.

    Currently does not describe a full area detector setup. It is just
    meant to provide the bare essentials.

    """

    class CameraGroup(PVGroup):
        acquire_rbv = pvproperty(value=0, name="Acquire_RBV", dtype=PvpropertyInteger)
        acquire = pvproperty(value=0, name="Acquire", dtype=PvpropertyInteger)
        acquire_busy = pvproperty(value=0, name="AcquireBusy", dtype=PvpropertyInteger)
        gain = pvproperty(value=10, name="Gain")
        gain_rbv = pvproperty(value=10, name="Gain_RBV")

    cam = SubGroup(CameraGroup, prefix="cam1:")


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="255idSimDet:", desc="haven.tests.ioc_area_detector test IOC"
    )
    ioc = AreaDetectorGroup(**ioc_options)
    run(ioc.pvdb, **run_options)



import json
import warnings
import logging
from typing import Optional, Mapping, Sequence

from pydm.widgets import PyDMEmbeddedDisplay
import haven

from firefly import display

# from .voltmeter import VoltmeterDisplay


log = logging.getLogger(__name__)


class VoltmetersDisplay(display.FireflyDisplay):
    _ion_chamber_displays = []

    def __init__(
        self,
        args: Optional[Sequence] = None,
        macros: Mapping = {},
        **kwargs,
    ):
        self.ion_chambers = list(haven.registry.findall(label="ion_chambers"))
        macros_ = macros.copy()
        if "SCALER" not in macros_.keys():
            macros_["SCALER"] = self.ion_chambers[0].scaler_prefix
        super().__init__(args=args, macros=macros_, **kwargs)

    def customize_ui(self):
        # Delete existing voltmeter widgets
        for idx in reversed(range(self.voltmeters_layout.count())):
            self.voltmeters_layout.takeAt(idx).widget().deleteLater()
        # Add embedded displays for all the ion chambers
        try:
            ion_chambers = self.ion_chambers
        except haven.exceptions.ComponentNotFound as e:
            warnings.warn(str(e))
            log.warning(e)
            ion_chambers = []
        scaler_prefix = "CPT NOT FOUND"
        self._ion_chamber_displays = []
        for ic in sorted(ion_chambers, key=lambda c: c.ch_num):
            # Create the display object
            disp = PyDMEmbeddedDisplay(parent=self)
            disp.macros = json.dumps({"IC": ic.name})
            disp.filename = "voltmeter.py"
            # Add the Embedded Display to the Results Layout
            self.voltmeters_layout.addWidget(disp)
            self._ion_chamber_displays.append(disp)

    def ui_filename(self):
        return "voltmeters.ui"



import logging
from functools import lru_cache

from apsbss import apsbss
from qtpy.QtGui import QStandardItemModel, QStandardItem
from qtpy.QtCore import Signal, Slot
import qtawesome as qta

import haven
from firefly import display

log = logging.getLogger(__name__)


class BssDisplay(display.FireflyDisplay):
    """A PyDM display for the beamline scheduling system (BSS)."""

    _proposal_col_names = ["ID", "Title", "Start", "End", "Users", "Badges"]
    _esaf_col_names = [
        "ID",
        "Title",
        "Start",
        "End",
        "Users",
        "Badges",
    ]
    _esaf_id: str = ""
    _psoporsal_id: str = ""

    # Signal
    proposal_changed = Signal()
    proposal_selected = Signal()
    esaf_changed = Signal()
    esaf_selected = Signal()

    def __init__(self, api=apsbss, args=None, macros={}, **kwargs):
        self.api = api
        super().__init__(args=args, macros=macros, **kwargs)
        # Load data models for proposals and ESAFs
        self.load_models()

    def customize_ui(self):
        super().customize_ui()
        icon = qta.icon("fa5s.arrow-right")
        self.ui.update_proposal_button.setIcon(icon)
        self.ui.update_proposal_button.clicked.connect(self.update_proposal)
        self.ui.update_esaf_button.setIcon(icon)
        self.ui.update_esaf_button.clicked.connect(self.update_esaf)

    @property
    @lru_cache()
    def proposals(self):
        config = haven.load_config()
        proposals = []
        for proposal in self.api.getCurrentProposals(config["bss"]["beamline"]):
            users = proposal["experimenters"]
            proposals.append(
                {
                    "Title": proposal["title"],
                    "ID": proposal["id"],
                    "Start": proposal["startTime"],
                    "End": proposal["endTime"],
                    "Users": ", ".join([usr["lastName"] for usr in users]),
                    "Badges": ", ".join([usr["badge"] for usr in users]),
                }
            )
        return proposals

    @property
    @lru_cache()
    def esafs(self):
        config = haven.load_config()
        esafs_ = []
        for esaf in self.api.getCurrentEsafs(config["bss"]["beamline"].split("-")[0]):
            users = esaf["experimentUsers"]
            esafs_.append(
                {
                    "Title": esaf["esafTitle"],
                    "ID": esaf["esafId"],
                    "Start": esaf["experimentStartDate"],
                    "End": esaf["experimentEndDate"],
                    "Users": ", ".join([usr["lastName"] for usr in users]),
                    "Badges": ", ".join([usr["badge"] for usr in users]),
                }
            )
        return esafs_

    def load_models(self):
        config = haven.load_config()
        # Create proposal model object
        col_names = self._proposal_col_names
        self.proposal_model = QStandardItemModel()
        self.proposal_model.setHorizontalHeaderLabels(col_names)
        # Load individual proposals
        proposals = self.proposals
        for proposal in proposals:
            items = [QStandardItem(str(proposal[col])) for col in col_names]
            self.proposal_model.appendRow(items)
        self.ui.proposal_view.setModel(self.proposal_model)
        # Create proposal model object
        col_names = self._esaf_col_names
        self.esaf_model = QStandardItemModel()
        self.esaf_model.setHorizontalHeaderLabels(col_names)
        # Load individual esafs
        esafs = self.esafs
        for esaf in esafs:
            items = [QStandardItem(str(esaf[col])) for col in col_names]
            self.esaf_model.appendRow(items)
        self.ui.esaf_view.setModel(self.esaf_model)
        # Connect slots for when proposal/ESAF is changed
        self.ui.proposal_view.selectionModel().currentChanged.connect(
            self.select_proposal
        )
        self.ui.esaf_view.selectionModel().currentChanged.connect(self.select_esaf)

    def select_proposal(self, current, previous):
        # Determine which proposal was selected
        id_col_idx = self._proposal_col_names.index("ID")
        new_id = current.siblingAtColumn(id_col_idx).data()
        self._proposal_id = new_id
        # Enable controls for updating the metadata
        self.ui.update_proposal_button.setEnabled(True)
        self.proposal_selected.emit()

    def update_proposal(self):
        new_id = self._proposal_id
        # Change the proposal in the EPICS record
        bss = haven.registry.find(name="bss")
        bss.proposal.proposal_id.set(new_id).wait()
        self.api.epicsUpdate(bss.prefix)
        # Notify any interested parties that the proposal has been changed
        self.proposal_changed.emit()

    def select_esaf(self, current, previous):
        # Determine which esaf was selected
        id_col_idx = self._esaf_col_names.index("ID")
        new_id = current.siblingAtColumn(id_col_idx).data()
        self._esaf_id = new_id
        # Enable controls for updating the metadata
        self.ui.update_esaf_button.setEnabled(True)
        self.esaf_selected.emit()

    def update_esaf(self):
        new_id = self._esaf_id
        # Change the esaf in the EPICS record
        bss = haven.registry.find(name="bss")
        bss.wait_for_connection()
        bss.esaf.esaf_id.set(new_id).wait(timeout=5)
        self.api.epicsUpdate(bss.prefix)
        # Notify any interested parties that the esaf has been changed
        self.esaf_changed.emit()

    def ui_filename(self):
        return "bss.ui"



from ophyd import Device, FormattedComponent as FCpt, EpicsMotor

from .instrument_registry import registry
from .._iconfig import load_config


@registry.register
class XYStage(Device):
    """An XY stage with two motors operating in orthogonal directions.

    Vertical and horizontal are somewhat arbitrary, but are expected
    to align with the orientation of a camera monitoring the stage.

    Parameters
    ==========

    pv_vert
      The suffix to the PV for the vertical motor.
    pv_horiz
      The suffix to the PV for the horizontal motor.
    """

    vert = FCpt(EpicsMotor, "{prefix}{pv_vert}", labels={"motors"})
    horiz = FCpt(EpicsMotor, "{prefix}{pv_horiz}", labels={"motors"})

    def __init__(
        self,
        prefix: str,
        pv_vert: str,
        pv_horiz: str,
        labels={"stages"},
        *args,
        **kwargs
    ):
        self.pv_vert = pv_vert
        self.pv_horiz = pv_horiz
        super().__init__(prefix, labels=labels, *args, **kwargs)


def load_stages(config=None):
    if config is None:
        config = load_config()
    for name, stage_data in config.get("stage", {}).items():
        XYStage(
            name=name,
            prefix=stage_data["prefix"],
            pv_vert=stage_data["pv_vert"],
            pv_horiz=stage_data["pv_horiz"],
        )



import subprocess
from pathlib import Path

from qtpy.QtCore import Signal
from pydm import Display


class FireflyDisplay(Display):
    caqtdm_ui_file: str = ""
    caqtdm_command: str = "/APSshare/bin/caQtDM -style plastique -noMsg"

    # Signals
    status_message_changed = Signal(str, int)

    def __init__(self, parent=None, args=None, macros=None, ui_filename=None, **kwargs):
        super().__init__(
            parent=parent, args=args, macros=macros, ui_filename=ui_filename, **kwargs
        )
        self.customize_device()
        self.customize_ui()

    def _open_caqtdm_subprocess(self, cmds, *args, **kwargs):
        """Launch a new subprocess and save it to self._caqtdm_process."""
        # Try to leave this as just a simple call to Popen.
        # It helps simplify testing
        self._caqtdm_process = subprocess.Popen(cmds, *args, **kwargs)

    def launch_caqtdm(self, macros={}, ui_file: str = None):
        """Launch a caQtDM window showing the window's panel."""
        if ui_file is None:
            ui_file = self.caqtdm_ui_file
        cmds = self.caqtdm_command.split()
        macro_str = ",".join(f"{key}={val}" for key, val in macros.items())
        cmds = [*cmds, "-macro", macro_str, ui_file]
        self._open_caqtdm_subprocess(cmds)

    def customize_device(self):
        pass

    def customize_ui(self):
        pass

    def show_message(self, message, timeout=0):
        """Display a message in the status bar."""
        self.status_message_changed.emit(str(message), timeout)

    def ui_filename(self):
        raise NotImplementedError

    def ui_filepath(self):
        path_base = Path(__file__).parent
        return path_base / self.ui_filename()



from haven.instrument.power_supply import load_power_supplies


def test_load_power_supplies(sim_registry):
    load_power_supplies()
    # Test that the device has the right configuration
    devices = list(sim_registry.findall(label="power_supplies"))
    assert len(devices) == 2  # 2 channels on the device
    device = devices[0]
    assert "NHQ01_ch" in device.name
    pv_names = [d.potential.pvname for d in devices]
    assert "ps_ioc:NHQ01:Volt2_rbv" in pv_names



import time

from epics import caget, caput
from haven.instrument.heater import CapillaryHeater, load_heaters


def test_capillary_device(ioc_ptc10):
    # Set up the device and ensure connections match the IOC we have
    device = CapillaryHeater(ioc_ptc10.prefix, name="capillary_heater")
    device.wait_for_connection()
    # Check for status objects checking temperature properly
    status = device.set(200)
    assert not status.done
    caput(ioc_ptc10.pvs["tc1_temperature"], 200)
    new_temp = device.readback.get(use_monitor=False)
    assert new_temp == 200


def test_load_heaters(ioc_ptc10):
    heaters = load_heaters()
    assert len(heaters) == 1
    heater = heaters[0]
    print(heater)
    heaters[0].wait_for_connection()
    assert heaters[0].name == "capillary_heater"



from typing import Optional, Sequence
import logging
import warnings
from itertools import chain

from ophyd import Component, ophydobj

from .. import exceptions
from .._iconfig import load_config
from ..typing import Detector


log = logging.getLogger(__name__)


__all__ = ["InstrumentRegistry", "registry"]


def is_iterable(obj):
    return (not isinstance(obj, str)) and hasattr(obj, "__iter__")


def remove_duplicates(items, key=None):
    unique_items = list()
    for item in items:
        val = item if key is None else key(item)
        if val not in unique_items:
            yield item
            unique_items.append(val)


class InstrumentRegistry:
    """A registry keeps track of devices, signals, etc that have been
    previously registered.

    This mimics the %wa bluesky magic behavior, except that devices
    can be registered outside of the main REPL loop.

    """

    components: Sequence

    def __init__(self):
        self.clear()

    def clear(self):
        """Remove the previously registered components."""
        self.components = []

    @property
    def component_names(self):
        return [c.name for c in self.components]

    @property
    def device_names(self):
        return [c.name for c in self.components if c.parent is None]

    def find(
        self,
        any_of: Optional[str] = None,
        *,
        label: Optional[str] = None,
        name: Optional[str] = None,
        allow_none: Optional[str] = False,
    ) -> Component:
        """Find registered device components matching parameters.

        The *any_of* keyword is a proxy for all the other
        keywords. For example ``findall(any_of="my_device")`` is
        equivalent to ``findall(name="my_device",
        label="my_device")``.

        The name provided to *any_of*, *label*, or *name* can also
        include dot-separated attributes after the device name. For
        example, looking up ``name="eiger_500K.cam.gain"`` will look
        up the device named "eiger_500K" then return the
        Device.cam.gain attribute.

        Parameters
        ==========
        any_of
          Search by all of the other parameters.
        label
          Search by the component's ``labels={"my_label"}`` parameter.
        name
          Search by the component's ``name="my_name"`` parameter.
        allow_none
          If false, missing components will raise an exception. If
          true, an empty list is returned if no registered components
          are found.


        Returns
        =======
        result
          A list of all the components matching the search parameters.

        Raises
        ======
        ComponentNotFound
          No component was found that matches the given search
          parameters.
        MultipleComponentsFound
          The search parameters matched with more than one registered
          component. Either refine the search terms or use the
          ``self.findall()`` method.

        """
        results = list(
            self.findall(any_of=any_of, label=label, name=name, allow_none=allow_none)
        )
        if len(results) == 1:
            result = results[0]
        elif len(results) > 1:
            raise exceptions.MultipleComponentsFound(
                f"Found {len(results)} components matching query "
                f"[any_of={any_of}, label={label}, name={name}]. "
                "Consider using ``findall()``. "
                f"{results}"
            )
        else:
            result = None
        return result

    def _findall_by_label(self, label, allow_none):
        # Check for already created ophyd objects (return as is)
        if isinstance(label, ophydobj.OphydObject):
            yield label
            return
        # Recursively get lists of components
        if is_iterable(label):
            for lbl in label:
                yield from self.findall(label=lbl, allow_none=allow_none)
        else:
            # Split off label attributes
            try:
                label, *attrs = label.split(".")
            except AttributeError:
                attrs = []
            try:
                for cpt in self.components:
                    if label in getattr(cpt, "_ophyd_labels_", []):
                        # Re-apply the dot-notation attributes
                        cpt_ = cpt
                        for attr in attrs:
                            cpt_ = getattr(cpt_, attr)
                        yield cpt_
            except TypeError:
                raise exceptions.InvalidComponentLabel(label)

    def _findall_by_name(self, name):
        # Check for already created ophyd objects (return as is)
        if isinstance(name, ophydobj.OphydObject):
            yield name
            return
        # Check for an edge case with EpicsMotor objects (user_readback name is same as parent)
        try:
            is_user_readback = name[-13:] == "user_readback"
        except TypeError:
            is_user_readback = False
        if is_user_readback:
            parentname = name[:-14].strip("_")
            yield self.find(name=parentname).user_readback
        elif is_iterable(name):
            for n in name:
                yield from self.findall(name=n)
        else:
            # Split off any dot notation parameters for later filtering
            try:
                name, *attrs = name.split(".")
            except AttributeError:
                attrs = []
            # Find the matching components
            for cpt in self.components:
                if cpt.name == name:
                    cpt_ = cpt
                    # Re-apply dot-notation filter
                    for attr in attrs:
                        cpt_ = getattr(cpt_, attr)
                    yield cpt_

    def findall(
        self,
        any_of: Optional[str] = None,
        *,
        label: Optional[str] = None,
        name: Optional[str] = None,
        allow_none: Optional[bool] = False,
    ) -> Sequence[Component]:
        """Find registered device components matching parameters.

        Combining search terms works in an *or* fashion. For example,
        ``findall(name="my_device", label="ion_chambers")`` will find
        all devices that have either the name "my_device" or a label
        "ion_chambers".

        The *any_of* keyword is a proxy for all the other keywords. For
        example ``findall(any_of="my_device")`` is equivalent to
        ``findall(name="my_device", label="my_device")``.

        The name provided to *any_of*, *label*, or *name* can also
        include dot-separated attributes after the device name. For
        example, looking up ``name="eiger_500K.cam.gain"`` will look
        up the device named "eiger_500K" then return the
        Device.cam.gain attribute.

        Parameters
        ==========
        any_of
          Search by all of the other parameters.
        label
          Search by the component's ``labels={"my_label"}`` parameter.
        name
          Search by the component's ``name="my_name"`` parameter.
        allow_none
          If false, missing components will raise an exception. If
          true, an empty list is returned if no registered components
          are found.

        Returns
        =======
        results
          A list of all the components matching the search parameters.

        Raises
        ======
        ComponentNotFound
          No component was found that matches the given search
          parameters.

        """
        results = []
        # If using *any_of*, search by label and name
        _label = label if label is not None else any_of
        _name = name if name is not None else any_of
        # Apply several filters against label, name, etc.
        if is_iterable(any_of):
            for a in any_of:
                results.append(self.findall(any_of=a, allow_none=allow_none))
        else:
            # Filter by label
            if _label is not None:
                results.append(self._findall_by_label(_label, allow_none=allow_none))
            # Filter by name
            if _name is not None:
                results.append(self._findall_by_name(_name))
        # Peek at the first item to check for an empty result
        results = chain(*results)
        try:
            first = next(results)
        except StopIteration:
            # No results were found
            if allow_none:
                results = []
            else:
                raise exceptions.ComponentNotFound(
                    f'Could not find components matching: label="{_label}", name="{_name}"'
                )
        else:
            # Stick the first entry back in the queue and yield it
            results = chain([first], results)
        return remove_duplicates(results)

    def __new__wrapper(self, cls, *args, **kwargs):
        # Create and instantiate the new object
        obj = super(type, cls).__new__(cls)
        obj.__init__(*args, **kwargs)
        # Register the new object
        self.register(obj)
        return obj

    def register(self, component: Detector) -> Detector:
        """Register a device, component, etc so that it can be retrieved later.

        If *component* is a class, then any instances created will
        automatically be registered. Else, *component* will be assumed
        to be an instance and will be registered directly.

        Returns
        =======
        component
          The same component as was provided as an input.

        """
        beamline_is_connected = load_config()["beamline"]["is_connected"]
        # Determine how to register the device
        if isinstance(component, type):
            # A class was given, so instances should be auto-registered
            component.__new__ = self.__new__wrapper
        else:  # An instance was given, so just save it in the register
            # Ignore any instances with the same name as a previous component
            # (Needed for some sub-components that are just readback
            # values of the parent)
            duplicate_components = [
                c for c in self.components if c.name == component.name
            ]
            # Check that we're not adding a duplicate component name
            is_duplicate = component.name in [c.name for c in self.components]
            if is_duplicate:
                msg = f"Ignoring components with duplicate name: '{component.name}'"
                log.debug(msg)
                return component
            # Test the connection to ensure the device is present
            if beamline_is_connected and component.parent is None:
                try:
                    component.wait_for_connection()
                except TimeoutError:
                    msg = f"Could not connect to device {component.name} ({component.prefix})"
                    log.warning(msg)
                    return component
            # Register this component
            self.components.append(component)
            # Recusively register sub-components
            sub_signals = getattr(component, "_signals", {})
            for cpt_name, cpt in sub_signals.items():
                self.register(cpt)
        return component


registry = InstrumentRegistry()



from apstools.devices.aps_undulator import ApsUndulator

from .instrument_registry import registry
from .._iconfig import load_config


def load_xray_sources(config=None):
    if config is None:
        config = load_config()
    # Determine the X-ray source type (undulator vs bending magnet)
    data = config["xray_source"]
    if data["type"] == "undulator":
        dev = ApsUndulator(data["prefix"], name="undulator", labels={"xray_sources"})
        registry.register(dev)



import numpy as np
import pandas as pd
from bluesky import plans as bp, plan_stubs as bps
from bluesky.callbacks.core import CollectThenCompute
from bluesky.preprocessors import subs_decorator

from haven import instrument


class AutoGainCallback(CollectThenCompute):
    def __init__(self, devices, voltage_limits=(0.5, 4.5), *args, **kwargs):
        self.devices = devices
        self.best_sens_levels = {}
        self.llim = min(voltage_limits)
        self.hlim = max(voltage_limits)
        super().__init__(*args, **kwargs)

    def compute(self):
        # Empty dictionaries to hold results
        voltages = {dev.name: [] for dev in self.devices}
        sens_levels = {dev.name: [] for dev in self.devices}
        self.best_sens_levels = {}
        # Extract data from the events list
        for event in self._events:
            data = event["data"]
            for dev in self.devices:
                voltages[dev.name].append(data[f"{dev.name}_volts"])
                sens_levels[dev.name].append(data[f"{dev.name}_sensitivity_sens_level"])
        # Turn data into pandas series
        for device in self.devices:
            series = pd.Series(voltages[device.name], index=sens_levels[device.name])
            if (series < self.llim).all():
                best_level = series.index.max()
            elif (series > self.hlim).all():
                best_level = series.index.min()
            else:
                in_range_series = series[(series > self.llim) & (series < self.hlim)]
                best_level = in_range_series.idxmax()
                self.best_sens_levels[device.name] = best_level


def auto_gain(devices="ion_chambers"):
    """A plan to automatically set the gain on an ion chamber."""
    if isinstance(devices, str):
        devices = instrument.registry.findall(any=devices)
    # Prepare devices and device range arguments
    scan_args = []
    hints = []
    for device in devices:
        hints.append(device.sensitivity.sens_level.name)
        hints.append(device.volts.name)
        limits = device.sensitivity.sens_level.limits
        scan_args.extend(
            [device.sensitivity.sens_level, np.arange(limits[0], limits[1] + 1)]
        )
    # Hinting to make the best effort callback work properly
    _md = {
        "hints": {"dimensions": [(hints, "primary")]},
    }
    # Prepare detectors (we need to include the sensitivity level)
    detectors = [(dev, dev.volts, dev.sensitivity.sens_level) for dev in devices]
    detectors = [det for sublist in detectors for det in sublist]
    # Save motor kinds to restore later
    kinds = {dev.name: dev.sensitivity.kind for dev in devices}
    # Set up the callback to do the calculating
    cb = AutoGainCallback(devices=devices)
    plan_func = subs_decorator(cb)(bp.list_scan)
    try:
        # Set motor kinds so they get tracked during acquisition
        for dev in devices:
            dev.sensitivity.kind = "normal"
        yield from plan_func(detectors, *scan_args, md=_md)
    finally:
        for dev in devices:
            dev.sensitivity.kind = kinds[dev.name]
    # Update the devices with the new values
    mv_args = []
    for dev in devices:
        mv_args.append(dev.sensitivity.sens_level)
        mv_args.append(cb.best_sens_levels.get(dev.name, None))
    # For some reason, just doing to mv plan once doesn't set the values.
    # Should investigate more
    is_done = False
    while not is_done:
        is_done = True
        for dev in devices:
            best_level = cb.best_sens_levels.get(
                dev.name, dev.sensitivity.sens_level.get().readback
            )
            if dev.sensitivity.sens_level.get().readback != best_level:
                is_done = False
        yield from bps.mv(*mv_args)



from bluesky import plan_stubs as bps
from typing import Literal, Union, Sequence

from ..instrument.instrument_registry import registry
from ..typing import Motor


def _set_shutters(
    shutters: Union[str, Sequence[Motor]], direction: Literal["open", "closed"]
):
    shutters = registry.findall(shutters)
    # Prepare the plan
    plan_args = [obj for shutter in shutters for obj in (shutter, direction)]
    plan = bps.mv(*plan_args)
    # Emit the messages
    yield from plan


def open_shutters(shutters: Union[str, Sequence[Motor]] = "shutters"):
    """A plan to open the shutters.

    By default, this plan is greedy and will open all shutters defined
    at the beamline. If only specific shutters should be opened, they
    can be passed either as Device objects or device names as the
    *shutters* argument.

    E.g.
      RE(open_shutters(["Shutter C"]))

    E.g.
      shutter = haven.instrument.shutter.Shutter(..., name="Shutter C")
      RE(open_shutters([shutter]))

    This plan will temporarily remove the default shutter-related
    suspenders from the haven run engine.

    """
    yield from _set_shutters(shutters, "open")


def close_shutters(shutters: Union[str, Sequence[Motor]] = "shutters"):
    """A plan to close some shutters.

    By default, this plan is lazy and requires any shutters to be
    passed explicitly. Shutters can be passed either as Device objects
    or device names as the *shutters* argument.

    E.g.
      RE(close_shutters(["Shutter C"]))

    E.g.
      shutter = haven.instrument.shutter.Shutter(..., name="Shutter C")
      RE(close_shutters([shutter]))

    This plan will temporarily remove the default shutter-related
    suspenders from the haven run engine.

    """
    yield from _set_shutters(shutters, "closed")



from bluesky.callbacks import CallbackBase
from tqdm import tqdm


class ProgressBar(CallbackBase):
    num_points: int = 0

    """A progress bar that tracks a scan."""

    def start(self, doc):
        self.num_points = doc["num_points"]
        self.plan_name = doc.get("plan_name", "Scanning")
        # Set up the tqdm bar
        self.pbar = tqdm(total=self.num_points, unit="pt", desc=self.plan_name)

    def event(self, doc):
        self.pbar.update(1)

    def stop(self, doc):
        self.pbar.close()



class NoEnergies(ValueError):
    """The plan parameters do not result in any energies being selected for scanning."""

    ...


class GainOverflow(RuntimeError):
    """The gain is trying to be set to a value that is not allowed."""

    ...


class ComponentNotFound(IndexError):
    """Registry looked for a component, but it wasn't registered."""

    ...


class MultipleComponentsFound(IndexError):
    """Registry looked for a single component, but found more than one."""

    ...


class InvalidComponentLabel(TypeError):
    """Registry looked for a component, but the label provided is not vlaid."""

    ...


class FileNotWritable(IOError):
    """Output file is available but does not have write intent."""

    ...


class XDIFilenameKeyNotFound(KeyError):
    """The format string for an XDI filename has unknown placeholders."""

    ...


class InvalidPV(ValueError):
    """A process variable path or similar is expected, but the path
    provided is not valid.

    """

    ...


class DocumentNotFound(RuntimeError):
    """An attempt was made to retrieve a document from the mongodb database,
    but the requested document was not available."""

    ...


class IOCTimeout(RuntimeError):
    """The IOC did not start within the allotted time."""

    ...


class UnknownDeviceConfiguration(ValueError):
    """The configuration for a device does not match the known options."""

    ...


class InvalidHarmonic(ValueError):
    """The requested harmonic is invalid for this insertion device."""

    ...


class SignalNotFound(KeyError):
    """The dataset is not present in the run."""

    ...


class EmptySignalName(ValueError):
    ...


class InvalidTransformation(TypeError):
    """The data cannot be transformed to a new y-data signal."""

    ...



import time

from unittest import mock
import pytest
from qtpy import QtWidgets

import haven

from firefly.main_window import FireflyMainWindow
from firefly.application import FireflyApplication
from firefly.voltmeter import VoltmeterDisplay
from firefly.voltmeters import VoltmetersDisplay


def test_device(qtbot, ffapp, sim_registry):
    window = FireflyMainWindow()
    ic = haven.IonChamber("", ch_num=1, name="my_ion_chamber", labels={"ion_chambers"})
    sim_registry.register(ic)
    display = VoltmeterDisplay(macros={"IC": "my_ion_chamber"})
    assert hasattr(display, "_device")
    assert isinstance(display._device, haven.IonChamber)


def test_scaler_prefix(qtbot, ffapp, sim_registry):
    """Make sure the scaler prefix gets passed in as a macro."""
    # Set up fake ion chamber
    window = FireflyMainWindow()
    ic = haven.IonChamber(
        "",
        scaler_prefix="255idcVME:scaler1",
        ch_num=1,
        name="my_ion_chamber",
        labels={"ion_chambers"},
    )
    sim_registry.register(ic)
    # Check the macros
    display = VoltmetersDisplay()
    assert display.macros()["SCALER"] == "255idcVME:scaler1"


def test_gain_button(qtbot, ffapp):
    # Fake ion chamber to make sure the gain was actually changed
    I0 = mock.MagicMock()
    assert not I0.increase_gain.called
    # Prepare the UI
    window = FireflyMainWindow()
    display = VoltmeterDisplay(device=I0)
    display._device = I0
    button = display.ui.gain_up_button
    # Check that the button is a button
    assert type(button) is QtWidgets.QPushButton
    assert hasattr(button, "clicked")
    # Click the button and check that the gain changed
    button.click()
    assert I0.increase_gain.called


def test_current_display(qtbot, ffapp):
    """Test the labels that show the voltage converted to current
    based on amplifier settings.

    """
    window = FireflyMainWindow()
    display = VoltmeterDisplay(macros={"IOC_VME": "40idc", "CHANNEL_NUMBER": 1})
    assert hasattr(display, "_ch_gain_value")
    # Emit signals for changing the amplifier gain value
    display._ch_gain_value.value_slot(3)  # 10
    display._ch_gain_unit.value_slot(2)  # µA/V
    display._ch_voltage.value_slot(2.23)
    # Check that the label is updated
    assert display.gain == 10
    assert display.gain_unit == "µA"
    assert display.ui.ion_chamber_current.text() == "(22.30 µA)"


def test_embedded_display_widgets(qtbot, sim_registry, ffapp):
    """Test the the voltmeters creates a new embedded display widget for
    each ion chamber.

    """
    window = FireflyMainWindow()
    # Set up fake ion chambers
    I0 = haven.IonChamber(
        prefix="eggs_ioc", ch_num=2, name="I0", labels={"ion_chambers"}
    )
    sim_registry.register(I0)
    It = haven.IonChamber(
        prefix="spam_ioc", ch_num=3, name="It", labels={"ion_chambers"}
    )
    sim_registry.register(It)
    # Load the display
    vms_display = VoltmetersDisplay()
    # Check that the embedded display widgets get added correctly
    assert hasattr(vms_display, "_ion_chamber_displays")
    assert len(vms_display._ion_chamber_displays) == 2
    assert vms_display.voltmeters_layout.count() == 2
    # Check that the embedded display widgets have the correct macros
    emb_disp = vms_display._ion_chamber_displays[0]
    disp = emb_disp.open_file(force=True)
    macros = disp.macros()
    assert macros == {"IC": "I0", "SCALER": "eggs_ioc"}
    # Check that a device has been created properly
    assert type(disp._device) is haven.IonChamber



import time

import epics

from haven import Monochromator


def test_mono_energy_signal(ioc_mono):
    mono = Monochromator(ioc_mono.prefix.strip(":"), name="monochromator")
    mono.wait_for_connection(timeout=20)
    time.sleep(0.1)
    # Change mono energy
    mono.energy.set(5000)
    # Check new value on the IOC
    assert epics.caget(ioc_mono.pvs["energy"], use_monitor=False) == 5000



import pytest
import json

import haven
from pydm.data_plugins.epics_plugin import EPICSPlugin
from pydm.widgets.channel import PyDMChannel
from qtpy import QtWidgets, QtGui, QtCore

from firefly.main_window import FireflyMainWindow
from firefly.cameras import CamerasDisplay
from firefly.camera import CameraDisplay, DetectorStates


macros = {"PREFIX": "camera_ioc:", "DESC": "Camera A"}


def test_embedded_displays(qtbot, ffapp, sim_registry, sim_camera):
    """Test that the embedded displays get loaded."""
    FireflyMainWindow()
    # Set up fake cameras
    # camera = haven.Camera(prefix="camera_ioc:", name="Camera A", labels={"cameras"})
    # sim_registry.register(camera)
    # Load the display
    display = CamerasDisplay()
    # Check that the embedded display widgets get added correctly
    assert hasattr(display, "_camera_displays")
    assert len(display._camera_displays) == 1
    # Check the embedded display macros
    # assert isinstance(display._camera_displays[0].macros, dict)
    expected_macros = {"CAMERA": sim_camera.name, "DESC": sim_camera.description}
    assert json.loads(display._camera_displays[0].macros) == expected_macros


def test_camera_channel_status(qtbot, ffapp):
    """Test that the camera status indicator responds to camera connection
    status PV.

    """
    FireflyMainWindow()
    display = CameraDisplay(macros=macros)
    # Check that the pydm connections have been made to EPICS
    assert isinstance(display.detector_state, PyDMChannel)
    assert display.detector_state.address == "camera_ioc:cam1:DetectorState_RBV"


def test_set_status_byte(qtbot, ffapp):
    FireflyMainWindow()
    display = CameraDisplay(macros=macros)
    display.show()
    # All devices are disconnected
    state = display.detector_state
    byte = display.camera_status_indicator
    bit = byte._indicators[0]
    label = display.camera_status_label
    # Set the color to something else, then check that it gets set back to white
    bit.setColor(QtGui.QColor(255, 0, 0))
    # Simulated the IOC being disconnected
    display.update_camera_connection(False)
    assert bit._brush.color().getRgb() == (255, 255, 255, 255)
    assert not label.isVisible(), "State label should be hidden by default"
    # Make the signals connected and see that it's green
    display.update_camera_connection(True)
    display.update_camera_state(DetectorStates.IDLE)
    assert bit._brush.color().getRgb() == (0, 255, 0, 255)
    assert not label.isVisible(), "State label should be hidden by default"
    # Make the camera be disconnected and see if it's red
    display.update_camera_state(DetectorStates.DISCONNECTED)
    assert bit._brush.color().getRgb() == (255, 0, 0, 255)
    assert label.isVisible(), "State label should be visible when disconnected"
    # Make the camera be acquiring and see if it's yellow
    display.update_camera_state(DetectorStates.ACQUIRE)
    assert bit._brush.color().getRgb() == (255, 255, 0, 255)
    assert not label.isVisible(), "State label should be hidden by default"


@pytest.mark.xfail
def test_camera_viewer_button(qtbot, ffapp, ioc_area_detector, mocker):
    action = QtWidgets.QAction(ffapp)
    ffapp.camera_actions.append(action)
    FireflyMainWindow()
    display = CameraDisplay(macros=macros)
    display.show()
    # Click the button
    btn = display.ui.camera_viewer_button
    with qtbot.waitSignal(action.triggered):
        qtbot.mouseClick(btn, QtCore.Qt.LeftButton)



import time
import pytest
from unittest.mock import MagicMock
import asyncio

from bluesky import RunEngine, plans as bp
from qtpy.QtCore import QThread
from qtpy.QtTest import QSignalSpy
from bluesky_queueserver_api.zmq import REManagerAPI

from firefly.queue_client import QueueClient
from firefly.application import REManagerAPI
from firefly.main_window import FireflyMainWindow


def test_setup(ffapp):
    ffapp.setup_window_actions()
    ffapp.setup_runengine_actions()
    api = MagicMock()
    FireflyMainWindow()
    ffapp.prepare_queue_client(api=api)


def test_setup2(ffapp):
    """Verify that multiple tests can use the app without crashing."""
    ffapp.setup_window_actions()
    ffapp.setup_runengine_actions()
    api = MagicMock()
    FireflyMainWindow()
    ffapp.prepare_queue_client(api=api)



import time

from epics import caget
import pytest

from haven.instrument.lerix import (
    RowlandPositioner,
    LERIXSpectrometer,
    load_lerix_spectrometers,
)
import haven


um_per_mm = 1000


def test_rowland_circle_forward():
    rowland = RowlandPositioner(
        name="rowland", x_motor_pv="", y_motor_pv="", z_motor_pv="", z1_motor_pv=""
    )
    # Check one set of values
    um_per_mm
    result = rowland.forward(500, 60.0, 30.0)
    assert result == pytest.approx(
        (
            500.0 * um_per_mm,  # x
            375.0 * um_per_mm,  # y
            216.50635094610968 * um_per_mm,  # z
            1.5308084989341912e-14 * um_per_mm,  # z1
        )
    )
    # Check one set of values
    result = rowland.forward(500, 80.0, 0.0)
    assert result == pytest.approx(
        (
            484.92315519647707 * um_per_mm,  # x
            0.0 * um_per_mm,  # y
            171.0100716628344 * um_per_mm,  # z
            85.5050358314172 * um_per_mm,  # z1
        )
    )
    # Check one set of values
    result = rowland.forward(500, 70.0, 10.0)
    assert result == pytest.approx(
        (
            484.92315519647707 * um_per_mm,  # x
            109.92315519647711 * um_per_mm,  # y
            291.6982175363274 * um_per_mm,  # z
            75.19186659021767 * um_per_mm,  # z1
        )
    )
    # Check one set of values
    result = rowland.forward(500, 75.0, 15.0)
    assert result == pytest.approx(
        (
            500.0 * um_per_mm,  # x
            124.99999999999994 * um_per_mm,  # y
            216.50635094610965 * um_per_mm,  # z
            2.6514380968122676e-14 * um_per_mm,  # z1
        )
    )
    # Check one set of values
    result = rowland.forward(500, 71.0, 10.0)
    assert result == pytest.approx(
        (
            487.7641290737884 * um_per_mm,  # x
            105.28431301548724 * um_per_mm,  # y
            280.42235703910393 * um_per_mm,  # z
            68.41033299999741 * um_per_mm,  # z1
        )
    )


@pytest.mark.xfail
def test_rowland_circle_inverse():
    rowland = RowlandPositioner(
        name="rowland", x_motor_pv="", y_motor_pv="", z_motor_pv="", z1_motor_pv=""
    )
    # Check one set of values
    result = rowland.inverse(
        x=500.0,  # x
        y=375.0,  # y
        z=216.50635094610968,  # z
        z1=1.5308084989341912e-14,  # z1
    )
    assert result == pytest.approx((500, 60, 30))
    # # Check one set of values
    # result = rowland.forward(500, 80.0, 0.0)
    # assert result == pytest.approx((
    #     484.92315519647707 * um_per_mm,  # x
    #     0.0 * um_per_mm,  # y
    #     171.0100716628344 * um_per_mm,  # z
    #     85.5050358314172 * um_per_mm,  # z1
    # ))
    # # Check one set of values
    # result = rowland.forward(500, 70.0, 10.0)
    # assert result == pytest.approx((
    #     484.92315519647707 * um_per_mm,  # x
    #     109.92315519647711 * um_per_mm,  # y
    #     291.6982175363274 * um_per_mm,  # z
    #     75.19186659021767 * um_per_mm,  # z1
    # ))
    # # Check one set of values
    # result = rowland.forward(500, 75.0, 15.0)
    # assert result == pytest.approx((
    #     500.0 * um_per_mm,  # x
    #     124.99999999999994 * um_per_mm,  # y
    #     216.50635094610965 * um_per_mm,  # z
    #     2.6514380968122676e-14 * um_per_mm,  # z1
    # ))
    # # Check one set of values
    # result = rowland.forward(500, 71.0, 10.0)
    # assert result == pytest.approx((
    #     487.7641290737884 * um_per_mm,  # x
    #     105.28431301548724 * um_per_mm,  # y
    #     280.42235703910393 * um_per_mm,  # z
    #     68.41033299999741 * um_per_mm,  # z1
    # ))


@pytest.mark.xfail
def test_rowland_circle_component(ioc_motor):
    lerix = LERIXSpectrometer("255idVME", name="lerix")
    lerix.wait_for_connection()
    # Set pseudo axes
    statuses = [
        lerix.rowland.D.set(500.0),
        lerix.rowland.theta.set(60.0),
        lerix.rowland.alpha.set(30.0),
    ]
    # [s.wait() for s in statuses]  # <- this should work, need to come back to it
    time.sleep(0.1)
    # Check that the virtual axes were set
    result = lerix.rowland.get(use_monitor=False)
    assert caget("255idVME:m1") == pytest.approx(500.0 * um_per_mm)
    assert result.x.user_readback == pytest.approx(500.0 * um_per_mm)
    assert caget("255idVME:m2") == pytest.approx(375.0 * um_per_mm)
    assert result.y.user_readback == pytest.approx(375.0 * um_per_mm)
    assert caget("255idVME:m3") == pytest.approx(216.50635094610968 * um_per_mm)
    assert result.z.user_readback == pytest.approx(216.50635094610968 * um_per_mm)
    assert caget("255idVME:m4") == pytest.approx(1.5308084989341912e-14 * um_per_mm)
    assert result.z1.user_readback == pytest.approx(1.5308084989341912e-14 * um_per_mm)


def test_load_lerix_spectrometers(sim_registry):
    load_lerix_spectrometers()
    lerix = sim_registry.find(name="lerix")
    assert lerix.name == "lerix"
    assert lerix.x.prefix == "255idVME:m1"
    assert lerix.y.prefix == "255idVME:m2"
    assert lerix.z.prefix == "255idVME:m3"
    assert lerix.z1.prefix == "255idVME:m4"



from typing import Optional, Sequence
import logging

import pymongo
from pydantic import BaseModel
import intake
from bson.objectid import ObjectId
from bluesky import plan_stubs as bps

from .instrument.instrument_registry import registry
from . import exceptions

import time
from datetime import datetime

log = logging.getLogger(__name__)


__all__ = [
    "save_motor_position",
    "list_motor_positions",
    "recall_motor_position",
    "list_current_motor_positions",
]


class MotorAxis(BaseModel):
    name: str
    readback: float
    offset: float = None

    def as_dict(self):
        return {"name": self.name, "readback": self.readback, "offset": self.offset}


class MotorPosition(BaseModel):
    name: str
    motors: Sequence[MotorAxis]
    uid: Optional[str] = None
    savetime: Optional[float] = None

    def save(self, collection):
        payload = {
            "name": self.name,
            "motors": [m.as_dict() for m in self.motors],
            "savetime": self.savetime,
        }
        item_id = collection.insert_one(payload).inserted_id
        return item_id

    @classmethod
    def load(Cls, document):
        # Create a MotorPosition object
        motor_axes = [
            MotorAxis(
                name=m["name"], readback=m["readback"], offset=m.get("offset", None)
            )
            for m in document["motors"]
        ]
        position = Cls(
            name=document["name"],
            motors=motor_axes,
            uid=str(document["_id"]),
            savetime=document.get("savetime"),
        )
        return position


def default_collection():
    catalog = intake.catalog.load_combo_catalog()["haven"]
    client = catalog._asset_registry_db.client
    collection = client.get_database().get_collection("motor_positions")
    return collection


# Prepare the motor positions
def rbv(motor):
    """Helper function to get readback value (rbv)."""
    try:
        # Wrap this in a try block because not every signal has this argument
        motor_data = motor.get(use_monitor=False)
    except TypeError:
        log.debug("Failed to do get() with ``use_monitor=False``")
        motor_data = motor.get()
    if hasattr(motor_data, "readback"):
        return motor_data.readback
    elif hasattr(motor_data, "user_readback"):
        return motor_data.user_readback
    else:
        return motor_data


def save_motor_position(*motors, name: str, collection=None):
    """Save the current positions of a number of motors to a database.

    Parameters
    ==========
    *motors
      The list of motors (or motor names/labels) whose position to
      save.
    name
      A human-readable name for this position (e.g. "sample center")
    collection
      A pymongo collection object to receive the data. Meant for
      testing.

    Returns
    =======
    item_id
      The ID of the item in the database.
    """
    # Get default collection if none was given
    if collection is None:
        collection = default_collection()
    # Resolve device names or labels
    motors = registry.findall(motors)
    # Prepare the motor positions
    motor_axes = []
    for m in motors:
        payload = dict(name=m.name, readback=rbv(m))
        # Save the calibration offset for motors
        if hasattr(m, "user_offset"):
            payload["offset"] = m.user_offset.get()
        axis = MotorAxis(**payload)
        motor_axes.append(axis)
    savetime = time.time()
    position = MotorPosition(name=name, motors=motor_axes, savetime=savetime)
    # Write to the database
    pos_id = position.save(collection=collection)
    log.info(f"Saved motor position {name} (uid={pos_id})")
    return pos_id


def print_motor_position(position):
    BOLD = "\033[1m"
    END = "\033[0m"
    # Prepare metadata strings for the header
    metadata = []
    if position.uid is not None:
        metadata.append(f'uid="{position.uid}"')
    if position.savetime is not None:
        timestamp = datetime.fromtimestamp(position.savetime).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        ts_str = f"timestamp={timestamp}"
        metadata.append(ts_str)
    if len(metadata) > 0:
        metadata_str = f" ({', '.join(metadata)})"
    else:
        metadata_str = ""
    # Write the output header
    output = f"\n{BOLD}{position.name}{END}{metadata_str}\n"
    # Write the motor positions
    for idx, motor in enumerate(position.motors):
        # Figure out some nice tree aesthetics
        is_last_motor = idx == (len(position.motors) - 1)
        box_char = "┗" if is_last_motor else "┣"
        output += f"{box_char}━{motor.name}: {motor.readback}, offset: {motor.offset}\n"
    print(output, end="")


def list_motor_positions(collection=None):
    """Print a list of saved motor positions.

    The name and UID will be printed, along with each motor and it's
    position.

    Parameters
    ==========
    collection
      The mongodb collection from which to print motor positions.

    """
    # Get default collection if none was given
    if collection is None:
        collection = default_collection()
    # Get the motor positions from disk
    results = collection.find()
    # Go through the results and display them
    were_found = False
    for doc in results:
        were_found = True
        position = MotorPosition.load(doc)
        print_motor_position(position)
    # Some feedback in the case of empty motor positions
    if not were_found:
        print(f"No motor positions found: {collection}")


def get_motor_position(
    uid: Optional[str] = None, name: Optional[str] = None, collection=None
) -> MotorPosition:
    """Retrieve a previously saved motor position from the database.

    Parameters
    ==========
    uid
      The universal identifier for the the document in the collection.
    name
      The name of the saved motor position, as given with the *name*
      parameter to the ``save_motor_position`` function.
    collection
      The mongodb collection from which to print motor positions.

    Returns
    =======
    position
      The most recent motor position with data retrieved from the
      database.

    """
    # Check that at least one of the parameters is given
    has_query_param = any([val is not None for val in [uid, name]])
    if not has_query_param:
        raise TypeError("At least one query parameter (*uid*, *name*) is required")
    # Get default collection if none was given
    if collection is None:
        collection = default_collection()
    # Build query for finding motor positions
    if uid is not None:
        _id = ObjectId(uid)
    else:
        _id = None
    query_params = {"_id": _id, "name": name}
    # Filter out query parameters that are ``None``
    query_params = {k: v for k, v in query_params.items() if v is not None}
    result = collection.find_one(query_params, sort=[("savetime", pymongo.DESCENDING)])
    # Feedback for if no matching motor positions are in the database
    if result is None:
        raise exceptions.DocumentNotFound(
            f'Could not find document matching: {query_params}"'
        )
    position = MotorPosition.load(result)
    return position


def recall_motor_position(
    uid: Optional[str] = None, name: Optional[str] = None, collection=None
):
    """Set motors to their previously saved positions.

    Parameters
    ==========
    uid
      The universal identifier for the the document in the collection.
    name
      The name of the saved motor position, as given with the *name*
      parameter to the ``save_motor_position`` function.
    collection
      The mongodb collection from which to print motor positions.

    """
    # Get default collection if none was given
    if collection is None:
        collection = default_collection()
    # Get the saved position from the database
    position = get_motor_position(uid=uid, name=name, collection=collection)
    # Create a move plan to recall the position
    plan_args = []
    for axis in position.motors:
        motor = registry.find(name=axis.name)
        plan_args.append(motor)
        plan_args.append(axis.readback)
    yield from bps.mv(*plan_args)


def list_current_motor_positions(*motors, name="current motor"):
    """list and print the current positions of a number of motors

    Parameters
    ==========
    *motors
      The list of motors (or motor names/labels) whose position to
      save.
    name
      A human-readable name for this position (e.g. "sample center")

    """
    # Resolve device names or labels
    motors = [registry.find(name=m) for m in motors]
    # Build the list of motor positions
    motor_axes = []
    for m in motors:
        payload = dict(name=m.name, readback=rbv(m))
        # Save the calibration offset for motors
        if hasattr(m, "user_offset"):
            payload["offset"] = m.user_offset.get()
        axis = MotorAxis(**payload)
        motor_axes.append(axis)
    position = MotorPosition(
        name=name, motors=motor_axes, uid=None, savetime=time.time()
    )
    # Display the current motor positions
    print_motor_position(position)



from unittest.mock import MagicMock
import pytest
from ophyd import sim
from bluesky.callbacks.best_effort import BestEffortCallback
from lmfit.models import QuadraticModel

from haven import mono_ID_calibration
from run_engine import RunEngineStub


@pytest.fixture()
def mono_motor(sim_registry):
    motor = sim_registry.register(
        sim.SynAxis(name="energy_mono_energy", labels={"motors", "energies"})
    )
    yield motor


@pytest.fixture()
def pitch2_motor(sim_registry):
    motor = sim_registry.register(
        sim.SynAxis(name="monochromator_pitch2", labels={"motors"})
    )
    yield motor


@pytest.fixture()
def id_motor(sim_registry):
    motor = sim_registry.register(
        sim.SynAxis(name="energy_id_energy", labels={"motors", "energies"})
    )
    yield motor


@pytest.fixture()
def ion_chamber(sim_registry, id_motor):
    I0 = sim_registry.register(
        sim.SynGauss(
            "I0",
            id_motor,
            "energy_id_energy",
            center=8.0,
            Imax=1,
            sigma=1,
            labels={"ion_chambers"},
        )
    )
    yield I0


def test_moves_energy(mono_motor, id_motor, ion_chamber, pitch2_motor):
    """Simple test to ensure that the plan moves the mono and undulators
    to the right starting energy."""
    # Execute the plan
    id_motor.set(8)
    fit_model = MagicMock()
    RE = RunEngineStub()
    RE(
        mono_ID_calibration(
            energies=[8000], energy_motor=mono_motor, fit_model=fit_model
        )
    )
    assert mono_motor.readback.get() == 8000.0


def test_aligns_pitch(mono_motor, id_motor, ion_chamber, pitch2_motor):
    fit_model = MagicMock()
    plan = mono_ID_calibration(
        energies=[8000], energy_motor=mono_motor, fit_model=fit_model
    )
    messages = list(plan)
    device_names = [getattr(m.obj, "name", None) for m in messages]
    assert "monochromator_pitch2" in device_names


def test_aligns_mono_energy(mono_motor, id_motor, ion_chamber, pitch2_motor):
    fit_model = MagicMock()
    plan = mono_ID_calibration(
        energies=[8000], energy_motor=mono_motor, fit_model=fit_model
    )
    messages = list(plan)
    id_messages = [
        m for m in messages if getattr(m.obj, "name", "") == "energy_id_energy"
    ]
    # Check that messages were produced for each point in the mono scan
    npts = 40  # Taken from `align_motor` plan
    assert len(id_messages) >= npts


def test_fitting_callback(mono_motor, id_motor, ion_chamber, pitch2_motor):
    fit_model = MagicMock()
    plan = mono_ID_calibration(
        energies=[8000, 9000], energy_motor=mono_motor, fit_model=fit_model
    )
    # Execute the plan in the runengine
    RE = RunEngineStub()
    result = RE(plan)
    # Check that the fitting results are available
    fit_model.fit.assert_called_once()
    # assert False



from haven.instrument.slits import load_slits


def test_load_slits(sim_registry):
    load_slits()
    # Check that the slits were loaded
    device = sim_registry.find(label="slits")
    assert device.prefix == "vme_crate_ioc:KB"
    assert device.h.center.setpoint.pvname == "vme_crate_ioc:KBHcenter"



import time
import pytest
from unittest.mock import MagicMock
import asyncio

from bluesky import RunEngine, plans as bp
from qtpy.QtCore import QThread
from qtpy.QtTest import QSignalSpy
from bluesky_queueserver_api.zmq import REManagerAPI

from firefly.queue_client import QueueClient
from firefly.application import REManagerAPI
from firefly.main_window import FireflyMainWindow


def test_setup(ffapp):
    ffapp.setup_window_actions()
    ffapp.setup_runengine_actions()
    api = MagicMock()
    FireflyMainWindow()
    ffapp.prepare_queue_client(api=api)


def test_queue_re_control(ffapp):
    """Test if the run engine can be controlled from the queue client."""
    api = MagicMock()
    ffapp.setup_window_actions()
    ffapp.setup_runengine_actions()
    ffapp.prepare_queue_client(api=api)
    window = FireflyMainWindow()
    window.show()
    # Try and pause the run engine
    ffapp.pause_runengine_action.trigger()
    # Check if the API paused
    time.sleep(0.1)
    api.re_pause.assert_called_once_with(option="deferred")
    # Pause the run engine now!
    api.reset_mock()
    ffapp.pause_runengine_now_action.trigger()
    # Check if the API paused now
    time.sleep(0.1)
    api.re_pause.assert_called_once_with(option="immediate")
    # Start the queue
    api.reset_mock()
    ffapp.start_queue_action.trigger()
    # Check if the queue started
    time.sleep(0.1)
    api.queue_start.assert_called_once()


def test_run_plan(ffapp, qtbot):
    """Test if a plan can be queued in the queueserver."""
    ffapp.setup_window_actions()
    ffapp.setup_runengine_actions()
    FireflyMainWindow()
    api = MagicMock()
    api.item_add.return_value = {"success": True, "qsize": 2}
    ffapp.prepare_queue_client(api=api)
    # Send a plan
    with qtbot.waitSignal(
        ffapp.queue_length_changed, timeout=1000, check_params_cb=lambda l: l == 2
    ):
        ffapp.queue_item_added.emit({})
    # Check if the API sent it
    api.item_add.assert_called_once_with(item={})


def test_check_queue_length(ffapp, qtbot):
    ffapp.setup_window_actions()
    ffapp.setup_runengine_actions()
    FireflyMainWindow()
    api = MagicMock()
    ffapp.prepare_queue_client(api=api)
    # Check that the queue length is changed
    api.get_queue.return_value = {
        "success": True,
        "msg": "",
        "items": [],
        "running_item": {},
        "plan_queue_uid": "f682e6fa-983c-4bd8-b643-b3baec2ec764",
    }
    with qtbot.waitSignal(
        ffapp.queue_length_changed, timeout=1000, check_params_cb=lambda l: l == 0
    ):
        ffapp._queue_client.check_queue_length()
    # Check that it isn't emitted a second time
    # with qtbot.waitSignal(ffapp.queue_length_changed):
    #     ffapp._queue_client.check_queue_length()
    # Now check a non-empty length queue
    api.queue_get.return_value = {
        "success": True,
        "msg": "",
        "items": ["hello", "world"],
        "running_item": {},
        "plan_queue_uid": "f682e6fa-983c-4bd8-b643-b3baec2ec764",
    }
    with qtbot.waitSignal(
        ffapp.queue_length_changed, timeout=1000, check_params_cb=lambda l: l == 2
    ):
        ffapp._queue_client.check_queue_length()



import warnings

import haven
from firefly import display


class MotorDisplay(display.FireflyDisplay):
    caqtdm_ui_file = "/APSshare/epics/synApps_6_2_1/support/motor-R7-2-2//motorApp/op/ui/autoconvert/motorx_all.ui"

    def customize_ui(self):
        self.debug_button.clicked.connect(self.launch_motor_caqtdm)

    def ui_filename(self):
        return "motor.ui"

    def launch_motor_caqtdm(self):
        warnings.warn("TODO: Set macros for launching motor window caqtdm.")
        caqtdm_macros = {}
        self.launch_caqtdm(macros=caqtdm_macros)



import time
from unittest import mock

import pytest
from qtpy import QtWidgets, QtCore
from bluesky_queueserver_api import BPlan
from apstools.devices.aps_undulator import ApsUndulator

import haven
from haven.instrument.monochromator import load_monochromator
from haven.instrument.energy_positioner import load_energy_positioner
from firefly.main_window import FireflyMainWindow
from firefly.energy import EnergyDisplay


def test_mono_caqtdm_macros(qtbot, ffapp, sim_registry):
    # Create fake device
    mono = haven.instrument.monochromator.Monochromator(
        "mono_ioc", name="monochromator"
    )
    sim_registry.register(
        haven.instrument.energy_positioner.EnergyPositioner(
            mono_pv="mono_ioc:Energy",
            id_offset_pv="mono_ioc:ID_offset",
            id_tracking_pv="mono_ioc:ID_tracking",
            id_prefix="id_ioc",
            name="energy",
        )
    )
    undulator = ApsUndulator("id_ioc:", name="undulator", labels={"xray_sources"})
    sim_registry.register(undulator)
    # Load display
    ffapp.setup_window_actions()
    ffapp.setup_runengine_actions()
    FireflyMainWindow()
    display = EnergyDisplay()
    display.launch_caqtdm = mock.MagicMock()
    # Check that the various caqtdm calls set up the right macros
    display.launch_mono_caqtdm()
    assert display.launch_caqtdm.called
    assert display.launch_caqtdm.call_args[1]["macros"] == {
        "P": "mono_ioc:",
        "MONO": "UP",
        "BRAGG": "ACS:m3",
        "GAP": "ACS:m4",
        "ENERGY": "Energy",
        "OFFSET": "Offset",
        "IDENERGY": "id_ioc:Energy",
    }


def test_id_caqtdm_macros(qtbot, ffapp, sim_registry):
    # Create fake device
    mono = haven.instrument.monochromator.Monochromator(
        "mono_ioc", name="monochromator"
    )
    sim_registry.register(
        haven.instrument.energy_positioner.EnergyPositioner(
            mono_pv="mono_ioc:Energy",
            id_offset_pv="mono_ioc:ID_offset",
            id_tracking_pv="mono_ioc:ID_tracking",
            id_prefix="id_ioc",
            name="energy",
        )
    )
    undulator = ApsUndulator("id_ioc:", name="undulator", labels={"xray_sources"})
    sim_registry.register(undulator)
    # Load display
    ffapp.setup_window_actions()
    ffapp.setup_runengine_actions()
    FireflyMainWindow()
    display = EnergyDisplay()
    display.launch_caqtdm = mock.MagicMock()
    # Check that the various caqtdm calls set up the right macros
    display.launch_id_caqtdm()
    assert display.launch_caqtdm.called
    assert display.launch_caqtdm.call_args[1]["macros"] == {
        "ID": "id_ioc",
        "M": 2,
        "D": 2,
    }


def test_move_energy(qtbot, ffapp, sim_registry):
    mono = haven.instrument.monochromator.Monochromator(
        "mono_ioc", name="monochromator"
    )
    sim_registry.register(
        haven.instrument.energy_positioner.EnergyPositioner(
            mono_pv="mono_ioc:Energy",
            id_offset_pv="mono_ioc:ID_offset",
            id_tracking_pv="mono_ioc:ID_tracking",
            id_prefix="id_ioc",
            name="energy",
        )
    )
    # Load display
    ffapp.setup_window_actions()
    ffapp.setup_runengine_actions()
    FireflyMainWindow()
    disp = EnergyDisplay()
    # Click the set energy button
    btn = disp.ui.set_energy_button
    expected_item = BPlan("set_energy", energy=8402.0)

    def check_item(item):
        return item.to_dict() == expected_item.to_dict()

    qtbot.keyClicks(disp.target_energy_lineedit, "8402")
    with qtbot.waitSignal(
        ffapp.queue_item_added, timeout=1000, check_params_cb=check_item
    ):
        qtbot.mouseClick(btn, QtCore.Qt.LeftButton)


def test_predefined_energies(qtbot, ffapp, ioc_mono, sim_registry):
    load_monochromator(
        config={
            "monochromator": {
                "ioc": "mono_ioc",
            },
        }
    )
    load_energy_positioner()
    # Create fake device
    mono = haven.instrument.monochromator.Monochromator(
        "mono_ioc", name="monochromator"
    )
    sim_registry.register(
        haven.instrument.energy_positioner.EnergyPositioner(
            mono_pv="mono_ioc:Energy",
            id_offset_pv="mono_ioc:ID_offset",
            id_tracking_pv="mono_ioc:ID_tracking",
            id_prefix="id_ioc",
            name="energy",
        )
    )
    # Set up the required Application state
    ffapp.setup_window_actions()
    ffapp.setup_runengine_actions()
    # Load display
    FireflyMainWindow()
    disp = EnergyDisplay()
    # Check that the combo box was populated
    combo_box = disp.ui.edge_combo_box
    assert combo_box.count() > 0
    assert combo_box.itemText(0) == "Select edge…"
    assert combo_box.itemText(1) == "Ca K (4038 eV)"
    # Does it filter energies outside the usable range?
    assert combo_box.count() < 250
    # Does it update the energy line edit?
    with qtbot.waitSignal(combo_box.activated, timeout=1000):
        qtbot.keyClicks(combo_box, "Ni K (8333 eV)\t")
        combo_box.activated.emit(9)  # <- this shouldn't be necessary
    line_edit = disp.ui.target_energy_lineedit
    assert line_edit.text() == "8333.000"



import time
import pytest
import logging

import epics
from haven.instrument import motor, registry
from qtpy import QtWidgets

from firefly.main_window import FireflyMainWindow
from firefly.application import FireflyApplication


def test_motor_menu(ioc_motor, sim_registry, qtbot, ffapp):
    prefix = ioc_motor.prefix.strip(":")
    motor.load_ioc_motors(prefix=prefix, num_motors=3)
    ffapp.setup_window_actions()
    ffapp.setup_runengine_actions()
    # Create the window
    window = FireflyMainWindow()
    # Check that the menu items have been created
    assert hasattr(window.ui, "menuPositioners")
    assert len(ffapp.motor_actions) == 3


def test_open_motor_window(sim_registry, ioc_motor, ffapp):
    # Set up motors in epics
    prefix = ioc_motor.prefix.strip(":")
    motor_name = "SLT_H_Inb"
    epics.caput(f"{prefix}:m1.DESC", motor_name, wait=True)
    motor.load_ioc_motors(prefix=prefix, num_motors=3)
    for m in sim_registry.findall(label="motors"):
        m.wait_for_connection()
    # Set up the application
    ffapp.setup_window_actions()
    ffapp.setup_runengine_actions()
    # Simulate clicking on the menu action (they're in alpha order)
    window = FireflyMainWindow()
    action = ffapp.motor_actions[2]
    action.trigger()
    # See if the window was created
    motor_1_name = "FireflyMainWindow_motor_SLT_H_Inb"
    assert motor_1_name in ffapp.windows.keys()
    macros = ffapp.windows[motor_1_name].display_widget().macros()
    assert macros["MOTOR"] == motor_name
    # Clean up
    window.close()



"""
Provide beamline configuration from the iconfig.toml file.

Example TOML configuration file: iconfig_default.toml

"""

__all__ = [
    "load_config",
]

import os
import logging
from typing import Sequence
import pathlib
import argparse
from functools import lru_cache

from mergedeep import merge
import tomli

log = logging.getLogger(__name__)


CONFIG_FILES = [
    pathlib.Path(__file__).parent / "iconfig_default.toml",
    pathlib.Path("~/bluesky/").expanduser() / "iconfig.toml",
    pathlib.Path("~/bluesky/instrument").expanduser() / "iconfig.toml",
]


def load_files(file_paths: Sequence[pathlib.Path]):
    """Generate the configs for files as dictionaries."""
    for fp in file_paths:
        if fp.exists():
            with open(fp, mode="rb") as fp:
                log.info(f"Loading config file: {fp}")
                config = tomli.load(fp)
                yield config

        else:
            log.debug(f"Could not find config file, skipping: {fp}")


@lru_cache()
def load_config(file_paths: Sequence[pathlib.Path] = CONFIG_FILES):
    """Load TOML config files.

    Will load files specified in *file_paths* and $HAVEN_CONFIG_FILES
    environmental variable.

    """
    file_paths = list(file_paths).copy()
    # Add config file from environmental variable
    try:
        file_paths.extend(
            [
                pathlib.Path(fp.strip())
                for fp in os.environ["HAVEN_CONFIG_FILES"].split(",")
            ]
        )
    except KeyError:
        pass
    # Load configuration from TOML files
    config = {}
    merge(config, *load_files(file_paths))
    return config


def print_config_value(args: Sequence[str] = None):
    """Print a config value from TOML files.

    Parameters
    ----------
    key
      The path to the value to retrieve from config files. Sections
      should be separated by dots, e.g. "shutter.A.open_pv"

    """
    # Set up command line arguments
    parser = argparse.ArgumentParser(
        prog="haven_config",
        description="Retrieve a value from Haven's config files.",
    )
    parser.add_argument("key", help="The dot-separated key to look up.")
    args = parser.parse_args(args=args)
    # Get the keys from the config file
    value = load_config()
    for part in args.key.split("."):
        value = value[part]
    print(value.strip())



#!/usr/bin/env python3
from caproto import ChannelType
from caproto.server import PVGroup, SubGroup, ioc_arg_parser, pvproperty, run

from ophyd.tests.scaler_ioc import EpicsScalerGroup


class PTC10Group(PVGroup):
    """
    An IOC for a PTC10 temperature controller.

    """

    class ThermocoupleGroup(PVGroup):
        temperature = pvproperty(
            value=21.3, doc="Temperature from thermocouple", name="temperature"
        )

    tc1 = SubGroup(ThermocoupleGroup, prefix="2A:")

    class PIDGroup(PVGroup):
        setpoint = pvproperty(value=0, doc="", name="setPoint")
        setpoint_rbv = pvproperty(value=0, doc="", name="setPoint_RBV")
        voltage = pvproperty(value=0, doc="", name="output")
        voltage_rbv = pvproperty(value=0, doc="", name="output_RBV")
        highlimit = pvproperty(value=0, doc="", name="highLimit")
        highlimit_rbv = pvproperty(value=0, doc="", name="highLimit_RBV")
        lowlimit = pvproperty(value=0, doc="", name="lowLimit")
        lowlimit_rbv = pvproperty(value=0, doc="", name="lowLimit_RBV")
        iotype = pvproperty(value=0, doc="", name="ioType")
        iotype_rbv = pvproperty(value=0, doc="", name="ioType_RBV")
        ramprate = pvproperty(value=0, doc="", name="rampRate")
        ramprate_rbv = pvproperty(value=0, doc="", name="rampRate_RBV")
        offswitch = pvproperty(value=0, doc="", name="off")
        pidmode = pvproperty(value=0, doc="", name="pid:mode")
        pidmode_rbv = pvproperty(value=0, doc="", name="pid:mode_RBV")
        P = pvproperty(value=0, doc="", name="pid:P")
        P_rbv = pvproperty(value=0, doc="", name="pid:P_RBV")
        I = pvproperty(value=0, doc="", name="pid:I")
        I_rbv = pvproperty(value=0, doc="", name="pid:I_RBV")
        D = pvproperty(value=0, doc="", name="pid:D")
        D_rbv = pvproperty(value=0, doc="", name="pid:D_RBV")
        inputchoice = pvproperty(value=0, doc="", name="pid:input")
        inputchoice_rbv = pvproperty(value=0, doc="", name="pid:input_RBV")
        tunelag = pvproperty(value=0, doc="", name="tune:lag")
        tunelag_rbv = pvproperty(value=0, doc="", name="tune:lag_RBV")
        tunestep = pvproperty(value=0, doc="", name="tune:step")
        tunestep_rbv = pvproperty(value=0, doc="", name="tune:step_RBV")
        tunemode = pvproperty(value=0, doc="", name="tune:mode")
        tunemode_rbv = pvproperty(value=0, doc="", name="tune:mode_RBV")
        tunetype = pvproperty(value=0, doc="", name="tune:type")
        tunetype_rbv = pvproperty(value=0, doc="", name="tune:type_RBV")

    pid1 = SubGroup(PIDGroup, prefix="5A:")


default_prefix = "ptc10ioc:"

if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="255idptc10:", desc="haven.tests.ioc_ptc10 test IOC"
    )
    ioc = PTC10Group(**ioc_options)
    run(ioc.pvdb, **run_options)



import pytest

from haven import registry, exceptions
from haven.instrument import stage


def test_stage_init():
    stage_ = stage.XYStage(
        "motor_ioc", pv_vert=":m1", pv_horiz=":m2", labels={"stages"}, name="aerotech"
    )
    assert stage_.name == "aerotech"
    assert stage_.vert.name == "aerotech_vert"
    # Check registry of the stage and the individiual motors
    registry.clear()
    with pytest.raises(exceptions.ComponentNotFound):
        registry.findall(label="motors")
    with pytest.raises(exceptions.ComponentNotFound):
        registry.findall(label="stages")
    registry.register(stage_)
    assert len(list(registry.findall(label="motors"))) == 2
    assert len(list(registry.findall(label="stages"))) == 1


def test_load_aerotech_stage():
    stage.load_stages()
    # Make sure these are findable
    stage_ = registry.find(name="Aerotech")
    assert stage_ is not None
    vert_ = registry.find(name="Aerotech_vert")
    assert vert_ is not None



import re
from typing import Callable, Union

from ophyd import Component, K


class RegexComponent(Component[K]):
    """A component with regular expression matching.

    In EPICS, it is not possible to add a field to an existing record,
    e.g. adding a ``.RnXY`` field to go alongside ``mca1.RnNM`` and
    other fields in the MCA record. A common solution is to create a
    new record with an underscore instead of the dot: ``mca1_RnBH``.

    This component include these types of field-like-records as part
    of the ROI device with a ``mca1.Rn`` prefix but performing
    subsitution on the device name using regular expressions. See the
    documentation for ``re.sub`` for full details.

    Example
    =======

    ```
    class ROI(mca.ROI):
        name = RECpt(EpicsSignal, "NM", lazy=True)
        is_hinted = RECpt(EpicsSignal, "BH",
                          pattern=r"^(.+)\.R(\d+)",
                          repl=r"\1_R\2",
                          lazy=True)

    class MCA(mca.EpicsMCARecord):
        roi0 = Cpt(ROI, ".R0")
        roi1 = Cpt(ROI, ".R1")

    mca = MCA(prefix="mca")
    # *name* has the normal concatination
    assert mca.roi0.name.pvname == "mca.R0NM"
    # *is_hinted* has regex substitution
    assert mca.roi0.is_hinted.pvname == "mca_R0BH"

    ```

    """

    def __init__(self, *args, pattern: str, repl: Union[str, Callable], **kwargs):
        """*pattern* and *repl* match their use in ``re.sub``."""
        self.pattern = pattern
        self.repl = repl
        super().__init__(*args, **kwargs)

    def maybe_add_prefix(self, instance, kw, suffix):
        """Parse prefix and suffix with regex suffix if kw is in self.add_prefix.

        Parameters
        ----------
        instance : Device
            The instance from which to extract the prefix to maybe
            append to the suffix.

        kw : str
            The key of associated with the suffix.  If this key is
            self.add_prefix than prepend the prefix to the suffix and
            return, else just return the suffix.

        suffix : str
            The suffix to maybe have something prepended to.

        Returns
        -------
        str

        """
        new_val = super().maybe_add_prefix(instance, kw, suffix)
        try:
            new_val = re.sub(self.pattern, self.repl, new_val)
        except TypeError:
            pass
        return new_val



import logging

import epics
from ophyd import EpicsMotor, EpicsSignal, Component as Cpt

from .._iconfig import load_config
from .instrument_registry import registry


log = logging.getLogger(__name__)


@registry.register
class HavenMotor(EpicsMotor):
    """The default motor for haven movement.

    Returns to the previous value when being unstaged.
    """

    description = Cpt(EpicsSignal, name="description", suffix=".DESC")
    tweak_value = Cpt(EpicsSignal, name="tweak_value", suffix=".TWV")
    tweak_forward = Cpt(EpicsSignal, name="tweak_forward", suffix=".TWF")
    tweak_reverse = Cpt(EpicsSignal, name="tweak_reverse", suffix=".TWR")
    soft_limit_violation = Cpt(EpicsSignal, name="soft_limit_violation", suffix=".LVIO")

    def stage(self):
        super().stage()
        # Save starting position to restore later
        self._old_value = self.user_readback.value

    def unstage(self):
        super().unstage()
        # Restore the previously saved position after the scan ends
        self.set(self._old_value, wait=True)


def load_all_motors(config=None):
    if config is None:
        config = load_config()
    # Figure out the prefix
    for name, config in config["motor"].items():
        prefix = config["prefix"]
        num_motors = config["num_motors"]
        log.info(f"Loading {num_motors} motors from IOC: {name} ({prefix})")
        load_ioc_motors(prefix=prefix, num_motors=num_motors, ioc_name=name)


def load_ioc_motors(prefix, num_motors, ioc_name=None):
    # Create motor objects
    motors = []
    for motor_num in range(num_motors):
        pv = f"{prefix}:m{motor_num+1}"
        # Take the name of the ophyd motor from the epics description
        log.debug(f"Looking for motor {motor_num} at {pv}")
        name = epics.caget(f"{pv}.DESC", timeout=1, use_monitor=False)
        if name is None:
            # The PV doesn't exist, so we've probably reached the highest motor number
            break
        # Create the motor device
        if name == f"motor {motor_num+1}":
            # It's an unnamed motor, so skip it
            log.debug(f"SKipping unnamed motor {motor_num}")
        else:
            # Create a new motor object
            labels = {"motors", "baseline"}
            if ioc_name is not None:
                labels = set([ioc_name, *labels])
            motor = HavenMotor(
                prefix=pv,
                name=name,
                labels=labels,
            )
            log.info(f"Created motor {motor}")
            motors.append(motor)



import pytest
import time

from haven.instrument.ion_chamber import (
    IonChamber,
    SensitivityLevelPositioner,
    load_ion_chambers,
)
from haven import exceptions
import epics


def test_gain_level(ioc_preamp, ioc_scaler):
    print(ioc_preamp)
    positioner = SensitivityLevelPositioner(
        f"{ioc_preamp.prefix}SR01", name="positioner"
    )
    positioner.wait_for_connection()
    assert positioner.get(use_monitor=False).sens_value.readback == epics.caget(
        ioc_preamp.pvs["preamp1_sens_num"]
    )
    assert positioner.get(use_monitor=False).sens_unit.readback == epics.caget(
        ioc_preamp.pvs["preamp1_sens_unit"]
    )
    # Move the gain level
    positioner.sens_level.set(12)
    time.sleep(0.1)  # Caproto breaks pseudopositioner status
    # Check that the preamp sensitivities are moved
    assert positioner.get(use_monitor=False).sens_value.readback == 3
    assert positioner.get(use_monitor=False).sens_unit.readback == 1
    # Check that the preamp sensitivity offsets are moved
    assert positioner.get(use_monitor=False).offset_value.readback == 0
    assert positioner.get(use_monitor=False).offset_unit.readback == 1
    # Change the preamp settings
    epics.caput(ioc_preamp.pvs["preamp1_sens_num"], 0)
    epics.caput(ioc_preamp.pvs["preamp1_sens_unit"], 3)
    epics.caput(ioc_preamp.pvs["preamp1_offset_num"], 1)
    epics.caput(ioc_preamp.pvs["preamp1_offset_unit"], 3)
    time.sleep(0.1)
    # Check that the gain level moved
    assert positioner.sens_level.get(use_monitor=False).readback == 27


def test_gain_changes(ioc_preamp, ioc_scaler):
    # Setup the ion chamber and connect to the IOC
    ion_chamber = IonChamber(
        prefix=ioc_scaler.prefix,
        preamp_prefix=f"{ioc_preamp.prefix}SR01",
        ch_num=2,
        name="ion_chamber",
    )
    time.sleep(0.01)
    ion_chamber.wait_for_connection(timeout=20)
    statuses = [
        ion_chamber.sensitivity.sens_value.set(2),
        ion_chamber.sensitivity.sens_unit.set(1),
    ]
    [status.wait() for status in statuses]
    assert ion_chamber.sensitivity.sens_value.get(use_monitor=False).readback == 2
    assert ion_chamber.sensitivity.sens_unit.get(use_monitor=False).readback == 1
    # Change the gain without changing units
    ion_chamber.increase_gain().wait()
    assert ion_chamber.sensitivity.sens_value.get(use_monitor=False).readback == 1
    assert ion_chamber.sensitivity.sens_unit.get(use_monitor=False).readback == 1
    ion_chamber.decrease_gain().wait()
    assert ion_chamber.sensitivity.sens_value.get(use_monitor=False).readback == 2
    assert ion_chamber.sensitivity.sens_unit.get(use_monitor=False).readback == 1
    # Change the gain so that it overflows and we have to change units
    max_sensitivity = len(ion_chamber.sensitivity.values) - 1
    max_unit = len(ion_chamber.sensitivity.units) - 1
    ion_chamber.sensitivity.sens_value.set(max_sensitivity).wait()
    assert ion_chamber.sensitivity.sens_value.get(use_monitor=False).readback == 8
    ion_chamber.decrease_gain().wait()
    assert ion_chamber.sensitivity.sens_value.get(use_monitor=False).readback == 0
    assert ion_chamber.sensitivity.sens_unit.get(use_monitor=False).readback == 2
    # Check that the gain can't overflow the acceptable values
    ion_chamber.sensitivity.sens_value.set(0).wait()
    ion_chamber.sensitivity.sens_unit.set(0).wait()
    with pytest.raises(exceptions.GainOverflow):
        ion_chamber.increase_gain()
    ion_chamber.sensitivity.sens_value.set(0).wait()
    ion_chamber.sensitivity.sens_unit.set(max_unit).wait()
    with pytest.raises(exceptions.GainOverflow):
        ion_chamber.decrease_gain()


def test_load_ion_chambers(sim_registry):
    load_ion_chambers()
    # Test the channel info is extracted properly
    ic = sim_registry.find(label="ion_chambers")
    assert ic.ch_num == 2
    assert ic.sensitivity.prefix.split(":")[-1] == "SR01"


def test_default_pv_prefix():
    """Check that it uses the *prefix* argument if no *scaler_prefix* is
    given.

    """
    prefix = "myioc:myscaler"
    # Instantiate the device with *scaler_prefix* argument
    device = IonChamber(
        name="device", prefix="gibberish", ch_num=1, scaler_prefix=prefix
    )
    device.scaler_prefix = prefix
    assert device.scaler_prefix == prefix
    # Instantiate the device with *scaler_prefix* argument
    device = IonChamber(name="device", ch_num=1, prefix=prefix)
    assert device.scaler_prefix == prefix


def test_offset_pv(sim_registry):
    """Check that the device handles the weird offset numbering scheme.

    Net count PVs in the scaler go as

    - 25idcVME:3820:scaler1_netA.B
    - 25idcVME:3820:scaler1_netA.C
    - etc.

    but the offset PVs go
    - 25idcVME:3820:scaler1_offset0.B
    - ...
    - 25idcVME:3820:scaler1_offset0.D
    - 25idcVME:3820:scaler1_offset1.A
    - ...

    """
    channel_suffixes = [
        (2, "offset0.B"),
        (3, "offset0.C"),
        (4, "offset0.D"),
        (5, "offset1.A"),
        (6, "offset1.B"),
        (7, "offset1.C"),
        (8, "offset1.D"),
        (9, "offset2.A"),
        (10, "offset2.B"),
        (11, "offset2.C"),
        (12, "offset2.D"),
    ]
    for ch_num, suffix in channel_suffixes:
        ic = IonChamber(
            prefix="scaler_ioc:scaler1", ch_num=ch_num, name=f"ion_chamber_{ch_num}"
        )
        assert ic.offset.pvname == f"scaler_ioc:scaler1_{suffix}", f"channel {ch_num}"



import display


class XafsScanRegionDisplay(display.FireflyDisplay):
    def ui_filename(self):
        return "xafs_scan_region.ui"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        obj = self.ui



import pytest

from haven.instrument.aps import load_aps, ApsMachine


def test_load_aps(sim_registry):
    load_aps()
    aps = sim_registry.find(name="APS")
    assert hasattr(aps, "current")


def test_read_attrs():
    device = ApsMachine(name="Aps")
    read_attrs = ["current", "lifetime"]
    for attr in read_attrs:
        assert attr in device.read_attrs


def test_config_attrs():
    device = ApsMachine(name="Aps")
    config_attrs = [
        "aps_cycle",
        "machine_status",
        "operating_mode",
        "shutter_permit",
        "fill_number",
        "orbit_correction",
        "global_feedback",
        "global_feedback_h",
        "global_feedback_v",
        "operator_messages",
    ]
    for attr in config_attrs:
        assert attr in device.configuration_attrs


def test_load_apsbss(sim_registry):
    load_aps()
    bss = sim_registry.find(name="bss")
    assert hasattr(bss, "esaf")
    assert hasattr(bss.esaf, "esaf_id")
    assert bss.esaf.esaf_id.pvname == "255idc:bss:esaf:id"
